/* --- Copyright University of Sussex 1995. All rights reserved. ----------
 > File:			C.unix/lib/auto/sys_open_url.p
 > Purpose:			Open an Internet URL
 > Author:			John Gibson, Dec  7 1995
 > Documentation:	REF * SOCKETS
 */
compile_mode :pop11 +strict;

uses unix_sockets;

section;

constant procedure sys_open_url;

lvars
	open_url,
	open_mode,
	open_org,
	open_org_rest,
	open_errc,
	open_url_rest
;

lconstant macro (
	DD_ATTR_LIST	= 1,
	DD_IODEV		= 2,
	DD_REM_NBYTES	= 3,
	DD_BUF			= 4,
	DD_BUF_NBYTES	= 5,
	DD_BUF_POS		= 6,
	DD_NET_ASCII	= 7,
	DD_VEC_LEN		= 7,
);

define lconstant Split_at(c, string, flags);
	lvars c, string, flags, n = locchar(c, 1, string);

	define lconstant decode(string);
		lvars string, i, len, c, d1, d2;

		define lconstant ishexcode(c);
			lvars c;
			if isnumbercode(c) then c fi_- `0`
			elseif `A` fi_<= c and c fi_<= `F` then c fi_- `7`
			elseif `a` fi_<= c and c fi_<= `f` then c fi_- `W`
			else false
			endif
		enddefine;

		returnunless(string and locchar(`%`, 1, string)) (string);
		datalength(string) -> len;
		consstring(#|
			fast_for i to len do
				if (fast_subscrs(i,string) ->> c) == `%` then
					i+2 -> i;
					if i fi_<= len
					and (ishexcode(fast_subscrs(i-1,string)) ->> d1)
					and (ishexcode(fast_subscrs(i,string)) ->> d2)
					then
						d1<<4 + d2
					else
						mishap(string, 1, 'INVALID "%" ENCODING IN URL')
					endif
				else
					c
				endif
			endfor
		|#)
	enddefine;

	if n then
		substring(1, n-1, string), allbutfirst(n, string)
	elseif flags &&/=_0 2:100 then
		string, false
	else
		false, string
	endif -> string;
	if flags &&/=_0 2:010 then decode() endif;
	string;
	if flags &&/=_0 2:001 then decode() endif
enddefine;

define Parse_url(url) -> (scheme, url);
	lvars url, n, scheme = false, url_path, user, port;
	Split_at(`:`, url, 2:000) -> (scheme, url);
	returnunless(scheme);
	uppertolower(scheme) -> scheme;
	returnunless(isstartstring('//', url));

	;;; Common IP scheme //<user>:<password>@<host>:<port>/<url-path>
	Split_at(`/`, allbutfirst(2, url), 2:100) -> (url, url_path);
	Split_at(`@`, url, 2:000) -> (user, url);
	Split_at(`:`, url, 2:110) -> (url, port);
	(port and (strnumber(port)->>port) and {^port tcp}) or scheme -> port;
	[%	(user and [% Split_at(`:`, user, 2:111) %]) :: [^url ^port],
		url_path or nullstring
	%] -> url
enddefine;


define lconstant Cons_dev(iodev, mode, buffer, net_ascii, close_p) -> dev;
	lvars iodev, mode, buffer, net_ascii, close_p, dev, name, fullname, dd;

	define lconstant Flush(dev);
		lvars dev;
		chain(subscrv(DD_IODEV,device_user_data(dev)), sysflush)
	enddefine;

	define lconstant Write(dev, bsub, userbuf, nbytes);
		lvars	n, len, bsub, pos, userbuf, lim, buf, nbytes, dev,
				dd = device_user_data(dev), iodev = subscrv(DD_IODEV,dd);
		unless fast_subscrv(DD_NET_ASCII,dd) then
			chain(iodev, bsub, userbuf, nbytes, fast_syswrite)
		endunless;

		;;; net ascii -- add c/r before newline
		fast_subscrv(DD_BUF,dd) -> buf;
		unless buf and datalength(buf) fi_>= nbytes fi_+ nbytes then
			inits(nbytes fi_+ nbytes fi_+ 512) ->> buf
												-> fast_subscrv(DD_BUF,dd)
		endunless;

		1 -> pos;
		bsub fi_+ nbytes -> lim;
		while bsub fi_< lim do
			unless (locchar(`\n`, bsub, userbuf) ->> n) and n fi_< lim then
				lim fi_- bsub -> len;
				move_subvector(bsub, userbuf, pos, buf, len);
				pos fi_+ len -> pos;
				quitloop
			endunless;
			n fi_- bsub -> len;
			if len == 1 then
				fast_subscrs(bsub,userbuf) -> fast_subscrs(pos,buf)
			elseif len /== 0 then
				move_subvector(bsub, userbuf, pos, buf, len)
			endif;
			pos fi_+ len -> pos;
			`\r` -> fast_subscrs(pos,buf);
			`\n` -> fast_subscrs(pos fi_+ 1,buf);
			pos fi_+ 2 -> pos;
			n fi_+ 1 -> bsub
		endwhile;
		chain(iodev, 1, buf, pos fi_- 1, fast_syswrite)
	enddefine;

	define lconstant Read(dev, bsub, userbuf, nbytes);
		lvars	dev, bsub, userbuf, nbytes, dd = device_user_data(dev),
				buf, buf_nbytes, iodev, len, pos, n, strip, rem_nbytes,
				org_nbytes = nbytes;

		returnif(nbytes == 0) (0);
		subscrv(DD_IODEV,dd) -> iodev;
		fast_subscrv(DD_BUF,dd) -> buf;
		if buf and (fast_subscrv(DD_BUF_NBYTES,dd) ->> buf_nbytes) == 0 then
			datalength(buf) -> len;
			if nbytes >= len then
				false -> buf
			else
				fast_sysread(iodev, 1, buf, len) -> buf_nbytes;
				1 -> subscrv(DD_BUF_POS,dd)
			endif
		endif;
		if fast_subscrv(DD_REM_NBYTES,dd) ->> rem_nbytes then
			fi_min(nbytes, rem_nbytes) -> nbytes
		endif;
		if buf then
			fast_subscrv(DD_BUF_POS,dd) -> pos;
			fi_min(buf_nbytes, nbytes) -> nbytes;
			if nbytes == 1 then
				fast_subscrs(pos,buf) -> fast_subscrs(bsub,userbuf)
			else
				move_subvector(pos, buf, bsub, userbuf, nbytes)
			endif;
			buf_nbytes fi_- nbytes -> fast_subscrv(DD_BUF_NBYTES,dd);
			pos fi_+ nbytes -> fast_subscrv(DD_BUF_POS,dd)
		else
			fast_sysread(iodev, bsub, userbuf, nbytes) -> nbytes
		endif;
		if rem_nbytes then
			rem_nbytes fi_- nbytes -> fast_subscrv(DD_REM_NBYTES,dd)
		endif;

		returnunless(fast_subscrv(DD_NET_ASCII,dd) ->> strip) (nbytes);

		;;; net ascii -- strip c/r before newline
		bsub fi_+ nbytes fi_- 1 -> len;		;;; subscr of last char in buffer
		if isinteger(strip) then
			;;; buffered char from last read
			true -> fast_subscrv(DD_NET_ASCII,dd);
			if nbytes == 0 then
				strip -> fast_subscrs(bsub,userbuf);
				return(1)
			elseunless strip == `\r`
			and fast_subscrs(bsub,userbuf) == `\n` then
				if nbytes == org_nbytes then
					;;; rebuffer the last char before shifting up
					fast_subscrs(len,userbuf) -> fast_subscrv(DD_NET_ASCII,dd);
					nbytes fi_- 1 -> nbytes
				endif;
				move_subvector(bsub, userbuf, bsub fi_+ 1, userbuf, nbytes);
				bsub fi_+ nbytes -> len;
				nbytes fi_+ 1 -> nbytes;
				strip -> fast_subscrs(bsub,userbuf)
			endif
		endif;

		bsub -> pos;
		while pos fi_<= len and (locchar(`\r`, pos, userbuf) ->> pos)
		and pos fi_<= len do
			fast_subscrv(DD_NET_ASCII,dd) -> strip;
			if pos == len then
				if isinteger(strip) then
					if strip == `\n` then
						`\n` -> fast_subscrs(pos,userbuf);
						true -> fast_subscrv(DD_NET_ASCII,dd)
					endif;
					return(nbytes)
				else
					;;; don't know next char -- buffer the c/r
					`\r` -> fast_subscrv(DD_NET_ASCII,dd);
					returnunless(nbytes == 1) (nbytes fi_- 1);
					chain(dev, bsub, userbuf, org_nbytes, Read)
				endif
			elseif fast_subscrs(pos fi_+ 1 ->> n,userbuf) == `\n` then
				;;; newline follows -- remove c/r
				if n == len then
					`\n` -> fast_subscrs(pos,userbuf)
				else
					move_subvector(n, userbuf, pos, userbuf, len fi_- n fi_+ 1)
				endif;
				if isinteger(strip) then
					;;; can unbuffer the char
					strip -> fast_subscrs(len,userbuf);
					true -> fast_subscrv(DD_NET_ASCII,dd)
				else
					nbytes fi_- 1 -> nbytes;
					len fi_- 1 -> len
				endif
			endif;
			n -> pos
		endwhile;
		return(nbytes)
	enddefine;		/* Read */

	define lconstant Test_input(dev);
		lvars dev, dd = device_user_data(dev), buf_nbytes = 0;
		if subscrv(DD_BUF,dd) then
			fast_subscrv(DD_BUF_NBYTES,dd) -> buf_nbytes
		endif;
		if isinteger(fast_subscrv(DD_NET_ASCII,dd)) then
			buf_nbytes fi_+ 1 -> buf_nbytes
		endif;
		if buf_nbytes /== 0 then
			buf_nbytes
		else
			sys_input_waiting(subscrv(DD_IODEV,dd))
		endif
	enddefine;

	define lconstant Clear_input(dev);
		lvars dev, dd = device_user_data(dev);
		0 -> subscrv(DD_BUF_NBYTES,dd);
		1 -> subscrv(DD_BUF_POS,dd);
		subscrv(DD_NET_ASCII,dd) and true -> subscrv(DD_NET_ASCII,dd);
		sys_clear_input(subscrv(DD_IODEV,dd))
	enddefine;

	lvars p_vec = {%
		mode == 0 and {^Read ^Test_input ^Clear_input},
		mode == 1 and {^Write ^Flush},
		false,
		close_p
	%};

	'URL\s' <> open_url -> name;
	consstring(#| explode('<URL:'), explode(open_url), `>` |#) -> fullname;
	initv(DD_VEC_LEN) -> dd;
	iodev		-> subscrv(DD_IODEV, dd);
	false		-> subscrv(DD_REM_NBYTES, dd);
	buffer		-> subscrv(DD_BUF, dd);
	0			-> subscrv(DD_BUF_NBYTES, dd);
	1			-> subscrv(DD_BUF_POS, dd);
	net_ascii	-> subscrv(DD_NET_ASCII, dd);
	[]			-> subscrv(DD_ATTR_LIST, dd);
	consdevice(name, fullname, dd, 0, p_vec) -> dev;
	sysclose -> sys_process_destroy_action(dev);
enddefine;

define lconstant Write_command(dev, cmnd, params);
	lvars dev, cmnd, params;
	consstring(#|
		explode(cmnd),
		if params /= nullstring then `\s`, explode(params) endif,
		`\r`, `\n`
	|#) -> cmnd;
	syswrite(dev, cmnd, datalength(cmnd))
enddefine;

define lconstant Get_status_code(s, len);
	lvars s, len, d1, d2, d3;

	define lconstant get_digit(c);
		lvars c;
		isnumbercode(c) and c fi_- `0`
	enddefine;

	len fi_>= 3
	and (get_digit(fast_subscrs(1,s)) ->> d1)
	and (get_digit(fast_subscrs(2,s)) ->> d2)
	and (get_digit(fast_subscrs(3,s)) ->> d3)
	and ((d1 fi_* 10 fi_+ d2) fi_* 10 fi_+ d3)
enddefine;

define lconstant Checkdl_rest(rest);
	lvars rest;
	if islist(rest) then
		dl(rest)
	else
		mishap(open_url, 1, 'INVALID URL (not Internet protocol format)')
	endif
enddefine;

define lconstant Open_ftp() -> dev;
	lvars	(hostspec, cdpath) = Checkdl_rest(open_url_rest),
			ftp_sock, dctrl_sock = false, d_sock = false, dev, dirlist,
			cwd, cwd_list, name, type, user, pass, resp, port, c, cmnd,
			mode = open_mode, errc = open_errc, localhost = sys_host_name();

	define lconstant close_devs();
		if ftp_sock then
			if d_sock then sysclose(d_sock) endif;
			if dctrl_sock then sysclose(dctrl_sock) endif;
			sysclose(ftp_sock)
		endif
	enddefine;

	dlocal 0 %	if dlocal_context == 1 then false -> ftp_sock endif,
				if dlocal_context == 2 then close_devs() endif
			 %;

	define lconstant ftp_mishap(mess);
		lvars mess;
		mishap(open_url, 1, 'ERROR OPENING ftp: URL (' sys_>< mess sys_>< ')')
	enddefine;

	define lconstant Get_response(cmnd, params);
		lvars cmnd, params, nread, b = sysstring;
		if cmnd then Write_command(ftp_sock, cmnd, params) endif;
		while (sysread(ftp_sock, b, sysstringlen) ->> nread) >= 4
		and fast_subscrs(4,b) == `-` and Get_status_code(b, nread) do
			;;; has continuation -- skip it
		endwhile;
		Get_status_code(b, nread)
	enddefine;


	;;; parse CWDs
	[%	while (Split_at(`/`, cdpath, 2:010) -> cdpath) ->> cwd do
			cwd
		endwhile
	%] -> cwd_list;
	Split_at(`;`, cdpath, 2:110) -> (name, type);
	if type then
		lowertoupper(type) -> type;
		unless datalength(type) == 6 and isstartstring('TYPE=',type)
		and strmember(subscrs(6,type)->>type, 'DLAI')
		then
			ftp_mishap('Invalid representation type')
		endunless
	endif;
	if name = nullstring and type /== `L` then `D` -> type endif;
	type == `D` or type == `L` -> dirlist;
	if dirlist and mode == 1 then
		ftp_mishap('Directory listing invalid in create mode')
	endif;

	;;; open connection
	sys_socket_to_service(tl(hostspec), "line") -> ftp_sock;
	unless Get_response(false, false) == 220 then
		ftp_mishap('Can\'t connect to ftp server')
	endunless;

	;;; login
	hd(hostspec) -> user;
	if user then dl(user) else 'anonymous', false endif -> (user, pass);
	unless (Get_response('USER', user) ->> resp) == 230 then
		if resp /== 331 then
			ftp_mishap('Login failed')
		elseif pass or user = 'anonymous' or user = 'ftp' then
			unless pass then popusername <> '@' <> localhost -> pass endunless;
			unless Get_response('PASS', pass) == 230 then
				ftp_mishap('Invalid username or password')
			endunless
		else
			ftp_mishap('Password required')
		endif
	endunless;

	;;; do cwds
	for cwd in cwd_list do
		unless (Get_response('CWD', cwd) ->> resp) == 250 then
			if resp == 550 then
				if errc == `A` or errc == `D` then chain(false, identfn) endif;
				ftp_mishap('Invalid directory')
			else
				ftp_mishap('Change directory failed')
			endif
		endunless
	endfor;

	;;; set representation type
	if open_org == true and not(type) then `I` -> type endif;
	if type and not(dirlist) then
		unless Get_response('TYPE', consstring(type, 1)) == 200 then
			ftp_mishap('Can\'t set representation type')
		endunless
	endif;

	if mode == 1 and errc == `F` and Get_response('SIZE', name) == 213 then
		;;; return false if exists
		chain(false, identfn)
	endif;

	;;; create control port and listen on it
	sys_socket(`i`, `S`, false) -> dctrl_sock;
	localhost -> sys_socket_name(dctrl_sock, 1);
	dl(sys_socket_name(dctrl_sock)) -> (localhost, port);
	consstring(#|
		fast_for c in_string localhost do
			if c == `.` then `,` else c endif
		endfor,
		`,`,
		dest_characters(port >> 8), `,`, dest_characters(port && 16:FF)
	|#) -> port;
	unless Get_response('PORT', port) == 200 then
		ftp_mishap('Can\'t set data port')
	endunless;

	;;; initiate transfer
	if type == `D` then
		'NLST'
	elseif type == `L` then
		'LIST'
	elseif mode == 0 then
		'RETR'
	else
		'STOR'
	endif -> cmnd;
	unless (Get_response(cmnd, name) ->> resp) == 150 then
		if resp = 550 then
			if errc /== `N` then chain(false, identfn) endif;
			'No such file'
		elseif resp == 553 then
			'Permission denied'
		else
			'Failed to open data connection'
		endif;
		ftp_mishap()
	endunless;

	sys_socket_accept(dctrl_sock, open_org) -> d_sock;
	sysclose(dctrl_sock);
	false -> dctrl_sock;

	define lconstant Close(dev);
		lvars dev, resp;
		if mode == 0 then
			until sysread(d_sock, sysstring, sysstringlen) == 0 do enduntil
		else
			sysflush(dev)
		endif;
		sysclose(d_sock);
		Get_response(false, false) -> resp;
		sysclose(ftp_sock);
		unless resp == 226 then ftp_mishap('Error in data transfer') endunless
	enddefine;

	Cons_dev(d_sock, mode, false, type/==`I`, Close) -> dev;
	if mode == 0 then
		[%	200, nullstring,
			'content-type:',if type == `I` then 'image/unknown'
							elseif dirlist then 'text/dirlist'
							else 'text/unknown'
							endif
		%] -> subscrv(DD_ATTR_LIST,device_user_data(dev))
	endif
enddefine;

define lconstant Open_http() -> dev;
	lvars	(hostspec, url_path) = Checkdl_rest(open_url_rest), sock,
			dev, dd, c, keywd, rest, resp, status, pos, content_length;

	define lconstant Close(dev);
		lvars dev;
		sysclose(sock)
	enddefine;

	define lconstant get_char(dev);
		lvars dev;
		lconstant b = writeable inits(1);
		if fast_sysread(dev, 1, b, 1) == 0 then
			termin
		else
			fast_subscrs(1,b)
		endif
	enddefine;

	define lconstant skip_white(dev) -> c;
		lvars dev, c;
		while (get_char(dev) ->> c) == `\s` or c == `\t` do endwhile
	enddefine;

	define lconstant get_header_line(c, dev, no_keywd) -> (c, keywd, rest);
		lvars dev, c, n = 0, rest, keywd = false, no_keywd;
		repeat
			unless c then get_char(dev) -> c endunless;
			if c == `\n` then
				returnunless(keywd or n /== 0) (false ->> rest -> c);
				quitunless((get_char(dev) ->> c) == `\s` or c == `\t`);
				skip_white(dev) -> c;
				`\s`; n fi_+ 1 -> n;
				nextloop
			endif;
			unless keywd or no_keywd then
				if c == `:` then
					consstring(c, n fi_+ 1) -> keywd;
					0 -> n;
					skip_white(dev) -> c;
					nextloop
				else
					uppertolower(c) -> c
				endif
			endunless;
			unless c == `\r` then c; n fi_+ 1 -> n endunless;
			false -> c
		endrepeat;
		consstring(n) -> rest
	enddefine;


	sys_socket_to_service(tl(hostspec), true) -> sock;
	Write_command(sock, 'GET', '/' <> url_path <> '\sHTTP/1.0');
	Write_command(sock, 'From:', popusername <> '@' <> sys_host_name());
	for c in open_org_rest do
		if isstring(c) then Write_command(sock, c, nullstring) endif
	endfor;
	Write_command(sock, nullstring, nullstring);

	Cons_dev(sock, 0, inits(512), false, Close) -> dev;
	device_user_data(dev) -> dd;

	unless get_char(dev) == `H` and get_char(dev) == `T`
	   and get_char(dev) == `T` and get_char(dev) == `P`
	   and get_char(dev) == `/`
	   and (;;; get status line
			repeat 3 times get_char(dev) -> endrepeat;
			get_header_line(skip_white(dev), dev, true) -> (c, , status);
			Get_status_code(status, datalength(status)) ->> resp
		)
	then
		;;; old server -- assume text/html
		subscrv(DD_BUF_POS,dd) -> pos;
		subscrv(DD_BUF_NBYTES,dd) + subscrv(DD_BUF_POS,dd) - 1
					-> subscrv(DD_BUF_NBYTES,dd);
		1 -> subscrv(DD_BUF_POS,dd);
		[200 '' 'content-type:' 'text/html'] -> subscrv(DD_ATTR_LIST,dd);
		return
	endunless;

	;;; read header lines
	false -> content_length;
	[%	resp, allbutfirst(3, status);
		while (get_header_line(c, dev, false) -> (keywd, rest)) ->> c do
			if keywd = 'content-length:' then
				strnumber(rest) -> content_length
			elseif keywd = 'location:' then
				'uri:' -> keywd
			endif;
			keywd, rest
		endwhile
	%] -> subscrv(DD_ATTR_LIST,dd);

	if resp == 200 or resp == 203 then
		;;; success or partial info -- return device
		if isinteger(content_length) then
			content_length -> subscrv(DD_REM_NBYTES,dd)
		endif
	else
		;;; anything else -- just return attribute list
		sysclose(sock);
		subscrv(DD_ATTR_LIST,dd) -> dev
	endif
enddefine;

define lconstant Open_file();
	lvars	(hostspec, url_path) = Checkdl_rest(open_url_rest),
			host = uppertolower(hd(tl(hostspec)));
	unless host = nullstring or host = 'localhost'
	or host = sys_host_name() then
		mishap(open_url, 1, 'INVALID file: URL (not local host)')
	endunless;
	unless isstartstring('$',url_path) or isstartstring('~',url_path) then
		'/' <> url_path -> url_path
	endunless;
	if open_mode == 0 then
		sysopen
	else
		syscreate
	endif (url_path, open_mode, open_org, open_errc)
enddefine;

define sys_open_url(url, mode, org);
	lvars url, mode, org, errc, scheme;

	if isinteger(org) then
		;;; optional character arg specifying error action
		((), url, mode, org) -> (url, mode, org, errc);
		unless errc == `N` or errc == `F`
		or (mode /== 1 and (errc == `D` or errc == `A`))
		then
			mishap(errc, 1, 'sys_open_url: INVALID ERROR CHARACTER CODE')
		endunless
	else
		if mode == 1 then `N` else `F` endif -> errc
	endif;

	check_string(url);
	checkinteger(mode, 0, 1);
	unless islist(org) then org :: [] -> org endunless;

	dlocal	open_url = url, open_mode = mode,
			(open_org, open_org_rest) = dest(org),
			open_errc = errc, open_url_rest;

	Parse_url(url) -> (scheme, open_url_rest);

	if scheme = 'ftp' then
		Open_ftp()

	elseif scheme = 'http' then
		Open_http()

	elseif scheme = 'file' then
		Open_file()

	else
		mishap(url, 1, 'INVALID URL (unknown or unsupported protocol)')
	endif;

enddefine;

endsection;
