/* --- Copyright University of Sussex 1995. All rights reserved. ----------
 * File:		C.x/x/Xpw/Text.c
 * Version:     Revision 5.0
 * Purpose: 	Contains XpwScrollTextWidget utility functions.
 * Author:		Jonathan Meyer (see revisions)
 * Date:		11 Nov 1989
 *
 */

/* using Xpw header files locally, not in <X11/Xpw/...> */
#define XpwNotInstalled

#include <stdlib.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>

#include "XpwScrTextP.h"

/* declarations */

extern
	_XpwTextCopyWait(),
	_XpwSetCharAttributes(),
	_XpwGetCharAttributes(),
	_XpwSetTextCursor(),
	_XpwGetTextCursor();

extern GC _XpwGetWorkTextGC();

XpwMethodRet
	_XpwTextMoveCursorTo();

static
	Clear();


/* The XpwScrollText widget has three core methods:
	Scroll - scrolls any rectangle of the screen in any direction.
	Clear  - clears a rectangle of the screen.
	Write  - writes an array of strings onto the screen in a rectangle.

All of the other methods are just specialisations of these four
*/

static void _XClearRectangle(dpy, win, gc, x, y, width, height)
Display *dpy;
Window win;
GC gc;
int x, y;
unsigned int width, height;
  { XGCValues gcv, newgcv;
	XGetGCValues(dpy, gc, GCForeground|GCBackground|GCFillStyle, &gcv);
	newgcv.foreground = gcv.background;
	newgcv.fill_style = FillSolid;
	XChangeGC(dpy, gc, GCForeground|GCFillStyle, &newgcv);
	XFillRectangle(dpy, win, gc, x, y, width, height);
	XChangeGC(dpy, gc, GCForeground|GCFillStyle, &gcv);
  }

void _XpwClearChars(w, col, row, ncols, nrows)
XpwScrollTextWidget w;
int row, col, nrows, ncols;
  { int sx, dx, fheight = FontHeight(w);
	Display *dpy = XtDisplay(w);
	Window win = XtWindow(w);

	ncols = min(ncols, NumCols(w) - col);
	nrows = min(nrows, NumRows(w) - row);
	if (ncols <= 0 || nrows <= 0) return;

	sx = ColToCoord(w, col);
	dx = ncols * FontWidth(w);

	if (row == 0)
	  { _XClearRectangle(dpy, win, w->xpwscrolltext.status_gc_array[0],
										sx, Row0Offset(w), dx, fheight);
		if (--nrows == 0) return;
	  }

	XClearArea(dpy, win, sx, MainRowToCoord(w,row), dx, nrows*fheight, FALSE);
  }


/*	Drawing a string -- also used by ScrollTextRedisplay
 */

static XpwScrollTextWidget blink_widget_chain = NULL;
static Boolean blink_on_phase = True, skip_cursor_blink = False;
static unsigned blink_row_start_col, blink_row_lim_col;
static XtIntervalId blink_timerid = (XtIntervalId)NULL;

static XpwScrollTextWidget onactive_widget = NULL;
static unsigned onactive_mode, onactive_row,
				onactive_start_col, onactive_lim_col;


#define DRAWLINE(x1,y1,x2,y2)	XDrawLine(dpy,win,gc,x1,y1,x2,y2)

#define DRAWSTRING(x,y,str,n)	\
	  { if (fontnum < 0) XDrawString(dpy,win,gc,x,y,(char*)str,n);		\
		else DrawString1By1(dpy,win,gc,x,y,(char*)str,n,				\
										FontStruct(w,fontnum),fwidth);	\
	  }

#define SETLSEG(X1,Y1,X2,Y2) \
		{ lsegp->x1=X1; lsegp->y1=Y1; lsegp->x2=X2; lsegp->y2=Y2; lsegp++; }


/*  Draw a string when the font width is proportional or less
	than FontWidth(w)
*/
static void DrawString1By1(dpy, win, gc, x, y, string, nchars, font, fwidth)
  Display *dpy;
  Window win;
  GC gc;
  int x, y;
  unsigned char *string;
  int nchars;
  XFontStruct *font;
  int fwidth;
  {	XTextItem items[MAX_COLS], *it = items, *itlim = it+MAX_COLS;
	XCharStruct *perchar = font->per_char, *pc = &font->max_bounds;
	int xoffs = 0, diff, leftsp, rightsp = 0;
	unsigned c, min_char = font->min_char_or_byte2,
				max_char = font->max_char_or_byte2;

	while (nchars--)
	  {	it->chars = (char *)string;
		it->nchars = 1;
		it->font = None;

		c = *string++;
		if (perchar)
		  {	if (c > max_char || c < min_char) c = font->default_char;
			pc = &perchar[c-min_char];
		  }

		diff = pc->rbearing - pc->lbearing;
		if (diff == 0
		/* some fonts seem to have the wrong rbearing for space */
		 || (c == ' ' && diff == 1))
			/* space */
			rightsp += fwidth;
		else
		  {	diff = fwidth - pc->width;
			leftsp = diff >> 1;
			it->delta = rightsp + leftsp;
			rightsp = diff - leftsp;
			it++;
		  }

		xoffs += fwidth;
		if (it == itlim || nchars == 0)
		  {	diff = it-items;
			if (diff != 0) XDrawText(dpy, win, gc, x, y, items, diff);
			it = items;
			x += xoffs;
			xoffs = rightsp = 0;
		  }
	  }
  }


/*	Draw graphics characters as codes 0-31 (vt100 encoding). Only the
	line-drawing ones and a few others are currently implemented --
	uses the font for the others.
*/
static void DrawGraphString(w, gc, xleft, ybase, string, nchars, bold, fontnum)
  XpwScrollTextWidget w;
  GC gc;
  int xleft, ybase;
  unsigned char *string;
  int nchars;
  Boolean bold;
  int fontnum;
  {
	Display *dpy = XtDisplay(w);
	Window win = XtWindow(w);
	int fwidth = FontWidth(w), fheight = FontHeight(w),
		fbaseline = FontBaseline(w), ytop = ybase - fbaseline,
		xmid = xleft+(fwidth-1)/2, ymid = ytop+fheight/2,
		xright = xleft+fwidth-1, ybot = ytop+fheight-1,
		d, r, h, v
		;
	XSegment lseg[6], *lsegp;
	XPoint pt[5];


	while (nchars--)
	  { switch (*string++)
		  { case 0:
				break;

			case 0x01:				/* diamond */
				h = (fbaseline*5)/8; if (!(h&1)) h++;

				pt[0].x = pt[2].x = xmid;
				pt[0].y = ybase-1;  pt[2].y = ybase-h;

				pt[3].x = xleft+(fwidth&1);
				pt[1].x = xright-1;
				pt[1].y = pt[3].y = ybase-1-h/2;

				pt[4].x = pt[0].x; pt[4].y = pt[0].y;
				XDrawLines(dpy, win, gc, pt, 5, CoordModeOrigin);
				XFillPolygon(dpy, win, gc, pt, 4, Convex, CoordModeOrigin);
				break;

			case 0x02:				/* checkerboard */
			case 0x03:				/* HT */
			case 0x04:				/* FF */
			case 0x05:				/* CR */
			case 0x06:				/* LF */
				goto usefont;

			case 0x07:				/* degree */
				d = (fwidth*5)/8;  r = d/2;
				v = xmid-r;
				h = fwidth; if (bold) h += 8;
				for (; h>0; h -= 8)
				  { XDrawArc(dpy, win, gc, v, ytop+fbaseline/3-r, d, d, 0, 360*64);
					v++;
				  }
				break;

			case 0x08:				/* plus or minus */
			case 0x09:				/* NL */
			case 0x0a:				/* VT */
				goto usefont;

			case 0x0b:				/* bottom-right ell */
				h = 1; v = 1; goto hvlines;

			case 0x0c:				/* top-right ell */
				h = 1; v = 2; goto hvlines;

			case 0x0d:				/* top-left ell */
				h = 2; v = 2; goto hvlines;

			case 0x0e:				/* bottom-left ell */
				h = 2; v = 1; goto hvlines;

			case 0x0f:				/* cross */
				h = 3; v = 3; goto hvlines;

			case 0x10:				/* horizontal line 1 */
				d = -2;  goto hline;

			case 0x11:				/* horizontal line 2 */
				d = -1;  goto hline;

			case 0x12:				/* horizontal line 3 (middle) */
				d = 0;  goto hline;

			case 0x13:				/* horizontal line 4 */
				d = 1;  goto hline;

			case 0x14:				/* horizontal line 5 */
				d = 2;  goto hline;

			case 0x15:				/* tee left */
				h = 2; v = 3; goto hvlines;

			case 0x16:				/* tee right */
				h = 1; v = 3; goto hvlines;

			case 0x17:				/* tee down */
				h = 3; v = 1; goto hvlines;

			case 0x18:				/* tee up */
				h = 3; v = 2; goto hvlines;

			case 0x19:				/* vertical line */
				h = 0; v = 3; goto hvlines;

			case 0x1a:				/* <= */
			case 0x1b:				/* >= */
			case 0x1c:				/* pi */
			case 0x1d:				/* /= */
			case 0x1e:				/* pound sign */
				goto usefont;

			case 0x1f:				/* centred dot */
				d = (fwidth+2)/4 - 1;
				d = max(1, d);
				h = xmid - d/2;
				if (!((fwidth|d) & 1)) h++;
				v = ybase-fbaseline/2+1;
				XDrawArc(dpy, win, gc, h, v, d, d, 0, 360*64);
				XFillArc(dpy, win, gc, h, v, d, d, 0, 360*64);
				break;


			/* Cursors */
			case ' ':				/* Transparent */
				break;

			case 'O':				/* Rect */
			block_cursor:
				XDrawRectangle(dpy, win, gc, xleft, ytop, fwidth-1, fheight-1);
				if (bold)
					XFillRectangle(dpy, win, gc, xleft, ytop, fwidth, fheight);
				break;

			case 'I':				/* I beam */
				lsegp = lseg;
				SETLSEG(xleft, ytop, xright, ytop);
				SETLSEG(xmid, ytop+1, xmid, ybot-1);
				SETLSEG(xleft, ybot, xright, ybot);
				if (bold)
				  { SETLSEG(xleft, ytop+1, xright, ytop+1);
					SETLSEG(xmid+1, ytop+2, xmid+1, ybot-2);
					SETLSEG(xleft, ybot-1, xright, ybot-1);
				  }
				XDrawSegments(dpy, win, gc, lseg, lsegp-lseg);
				break;

			case '^':				/* carat */
				pt[0].x = xright; pt[2].x = xleft;
				pt[0].y = pt[2].y = ybot;

				pt[1].x = xmid;
				pt[1].y = ybase-(fbaseline/3);

				pt[3].x = pt[0].x; pt[3].y = pt[0].y;
				XDrawLines(dpy, win, gc, pt, 4, CoordModeOrigin);
				if (bold)
					XFillPolygon(dpy, win, gc, pt, 3, Convex, CoordModeOrigin);
				break;

			case '_':				/* underscore */
				h = ybot-ybase+1;
				XDrawRectangle(dpy, win, gc, xleft, ybase, fwidth-1, h-1);
				if (bold)
					XFillRectangle(dpy, win, gc, xleft, ybase, fwidth, h);
				break;


			default:
				/* if unknown cursor, default to block */
				if (string[-1] >= 0x20) goto block_cursor;
				break;


			usefont:
				DRAWSTRING(xleft, ybase, string-1, 1);
				if (bold) DRAWSTRING(xleft+1, ybase, string-1, 1);
				break;

			hline:
				lsegp = lseg;
				d = ymid + d*((fheight+2)/5);
				SETLSEG(xleft, d, xright, d);
				if (bold) SETLSEG(xleft, d-1, xright, d-1);
				XDrawSegments(dpy, win, gc, lseg, lsegp-lseg);
				break;

			hvlines:
				lsegp = lseg;
				if (h != 0)
				  { int x1 = h==2?xmid:xleft, x2 = h==1?xmid:xright;
					SETLSEG(x1, ymid, x2, ymid);
					if (bold) SETLSEG(x1, ymid-1, x2, ymid-1);
				  }
				if (v != 0)
				  { int y1 = v==2?ymid:ytop, y2 = v==1?ymid:ybot;
					SETLSEG(xmid, y1, xmid, y2);
					if (bold) SETLSEG(xmid+1, y1, xmid+1, y2);
				  }
				XDrawSegments(dpy, win, gc, lseg, lsegp-lseg);
				break;
		  }

		xleft += fwidth; xmid += fwidth; xright += fwidth;
	  }
  }

#define DS_INVERT		1
#define DS_NOBACK		2
#define DS_ISCURSOR		4
#define DS_CURSORCHAR	8

#define ACT_LEAVE_NOTIFY	1
#define ACT_INVERT			2
#define ACT_BOX				4
#define ACT_POINTER2		8

static void DrawString(w, string, col, row, nchars, attrs, flags)
  XpwScrollTextWidget w;
  unsigned char *string;
  int col, row, nchars, attrs, flags;
  {
	Display
		*dpy = XtDisplay(w);
	Window
		win = XtWindow(w);
	int
		x = ColToCoord(w,col), y = RowToCoord(w,row),
		ybase = y + FontBaseline(w),
		gc_index = attrs & ATTR_GC_MASK, fontnum = gc_index%NFONTS,
		start_x = x, fwidth = FontWidth(w), indexlim, gcv_mask,
		colornum, startcol = col, actmode = 0, fontmode;
	Boolean
		bold = False, overstrike = False,
		draw_graph = w->xpwscrolltext.draw_graphic_chars,
		invert = flags & DS_INVERT,
		noback = flags & DS_NOBACK,
		cursorchar = flags & DS_CURSORCHAR,
		draw_shadow = False;
	GC
		*gc_array = w->xpwscrolltext.gc_array, gc;
	XGCValues
		gcv;
	Pixel
		shad_fg, shad_bg;
	register unsigned char
		*p = string, *plim = p+nchars;

	if (!nchars) return;

	if (attrs & ATTR_ACTIVE)
	  {	if (w == onactive_widget && row == onactive_row
		 && onactive_start_col <= startcol && startcol < onactive_lim_col)
			actmode = onactive_mode;
		/* treat as colors 8 - 15 except when cursor color */
		if (!(flags & DS_ISCURSOR)) gc_index |= ATTR_ACTIVE;
	  }

	colornum = gc_index/NFONTS;

	if (w->xpwscrolltext.draw_shadow_mask & (1<<colornum))
	  {	if (Flags(w) & MONOCHROME)
			/* use underline instead */
			attrs |= ATTR_UNDERLINE;
		else
		  {	unsigned s = row==0 ? 0 : 1;
			draw_shadow = True;
			gc_index -= fontnum; fontnum &= ~BOLDFONTBIT; gc_index += fontnum;
			shad_fg = BlackPixelOfScreen(XtScreen(w));
			shad_bg = w->xpwscrolltext.shadow_whites[s];
			if (w->xpwscrolltext.shadow_white_bits[s] & (1<<colornum))
			  { Pixel fg = shad_fg; shad_fg = shad_bg; shad_bg = fg; }
		  }
	  }

	gc = gc_array[fontnum];

	if (fontnum & BOLDFONTBIT)
	  { bold = True;
		if (!gc)
		  { /* bold mode with no bold font -- use non-bold font and
				overstrike */
			overstrike = True;
			gc_index -= fontnum;
			fontnum &= ~BOLDFONTBIT;
			gc_index += fontnum;
			gc = gc_array[fontnum];
		  }
	  }

	if (!gc || cursorchar)
	  { /* alt font mode with no alt font -- underline instead */
		if (cursorchar) attrs &= ~ATTR_UNDERLINE;
		else if (!draw_shadow) attrs |= ATTR_UNDERLINE;
		gc_index -= fontnum;
		fontnum = NORMAL_FONT;
		gc_index += NORMAL_FONT;
	  }

	/* get gc proper */
	if ((flags & DS_ISCURSOR) && colornum == 0)
	  { /* drawing cursor in colour 0 -- substitute cursor colour */
		gc_array = (row == 0) ? &w->xpwscrolltext.status_cursor_gc
							  : &w->xpwscrolltext.cursor_gc;
		indexlim = 1;
		colornum = CURSOR_COLORNUM;
	  }
	else
	  { if (row == 0)
		  {	gc_array = w->xpwscrolltext.status_gc_array;
			indexlim = NSTATUSGCS;
		  }
		else
			indexlim = NNORMALGCS;
	  }

	gc = gc_index < indexlim ? gc_array[gc_index]
							/* use work_gc and alter it */
							 : _XpwGetWorkTextGC(w, colornum, fontnum, row==0);



	/* draw the characters */

#define CLEARCHARS(n)													\
	  { if (!invert && colornum == 0)									\
			_XpwClearChars(w, col, row, n, 1);							\
		else															\
			_XClearRectangle(dpy, win, gc, x, y, n*fwidth, FontHeight(w));	\
	  }

	if (actmode & ACT_INVERT) invert = !invert;

	if ((attrs & ATTR_BLINK) && !w->xpwscrolltext.no_blink)
	  { /* set these values for DoBlink */
		blink_row_start_col = min(col, blink_row_start_col);
		blink_row_lim_col = max(col+nchars, blink_row_lim_col);
		if (!blink_on_phase)
		  { /* blinking chars and currently off -- clear them */
			if (!noback) CLEARCHARS(nchars);
			return;
		  }
	  }

	if (invert || draw_shadow)
	  { Pixel fg;
		gcv_mask = GCForeground | GCBackground;
		XGetGCValues(dpy, gc, gcv_mask, &gcv);
		if (invert)
		  { fg = gcv.foreground;
			gcv.foreground = gcv.background; gcv.background = fg;
			XChangeGC(dpy, gc, gcv_mask, &gcv);
			fg = shad_fg; shad_fg = shad_bg; shad_bg = fg;
		  }
	  }

	fontmode = (w->xpwscrolltext.font_draw_modes >> (fontnum<<1)) & 3;
	if (fontmode != 2) fontnum = -1;

	while (p < plim)
	  { int n;
		if (!cursorchar)
		  { if (draw_graph)
				do if (*p++ < 0x20) { --p; break; } while (p < plim);
			else
				p = plim;
		  }

		if (n = p - string)
		  { /* some non-graphics chars */
			Boolean draw = True;
			if (draw_shadow)
			  {	if (!noback) CLEARCHARS(n);
				XSetForeground(dpy, gc, shad_fg);
				DRAWSTRING(x+1, ybase+1, string, n);
				XSetForeground(dpy, gc, gcv.foreground);
			  }
			else if (!noback)
			  {	if (fontmode != 0)
					CLEARCHARS(n)
				else
				  {	XDrawImageString(dpy, win, gc, x, ybase, (char *)string, n);
					draw = False;
				  }
			  }
			if (draw) DRAWSTRING(x, ybase, string, n);
			if (overstrike) DRAWSTRING(x+1, ybase, string, n);
			string = p;
			x += n*fwidth;
			col += n;
		  }

		if (++p <= plim)
		  { /* some graphics chars */
			while (p < plim) if (*p++ >= 0x20) { --p; break; };
			n = p - string;
			if (!noback) CLEARCHARS(n);
			if (draw_shadow)
			  {	XSetForeground(dpy, gc, shad_fg);
				DrawGraphString(w, gc, x+1, ybase+1, string, n, bold, fontnum);
				XSetForeground(dpy, gc, gcv.foreground);
			  }
			DrawGraphString(w, gc, x, ybase, string, n, bold, fontnum);
			string = p;
			x += n*fwidth;
			col += n;
		  }
	  }

	/* underline */
	if (attrs & ATTR_UNDERLINE)
		DRAWLINE(start_x, ybase+1, x-1, ybase+1);

	if (actmode & ACT_BOX)
	  {	int ybot = y+FontHeight(w)-1, xright = x-1;
		XSegment lseg[4], *lsegp = lseg;
		SETLSEG(start_x, y,    xright, y);
		SETLSEG(start_x, ybot, xright, ybot);
		if (startcol == onactive_start_col) SETLSEG(start_x, y, start_x, ybot);
		if (col == onactive_lim_col)		SETLSEG(xright,  y, xright,  ybot);
		XDrawSegments(dpy, win, gc, lseg, lsegp-lseg);
	  }

	if (invert)
	  { Pixel fg = gcv.background;
		gcv.background = gcv.foreground; gcv.foreground = fg;
		XChangeGC(dpy, gc, gcv_mask, &gcv);
	  }
  }

void _XpwDrawRowSequences(w, scol, lcol, row, attr_mask)
XpwScrollTextWidget w;
register unsigned int scol, lcol, row;
unsigned int attr_mask;
  {
	WinRow rp = RowArray(w)[row];
	register WinChar wc, attrs = 0;
	unsigned char tmpstring[MAX_COLS];
	register unsigned char *outp = tmpstring;
	register int col, len = 0, coffs = COFFS(rp);

	col = coffs+scol;
	lcol += coffs;
	lcol = min(NumCols(w), lcol);

	while (col <= lcol)
	  { if (col == lcol || (wc = rp[col]) == 0 || (wc & 0xff00) != attrs)
		  { if (len != 0)
			  { DrawString(w, tmpstring, (col-coffs)-len, row, len, attrs>>8, 0);
				outp = tmpstring;
				len = 0;
			  }
			attrs = wc & 0xff00;
			if (col == lcol)
				break;
			else if (wc == 0)
			  { col++; continue; }
		  }

		if (!attr_mask || (attrs & attr_mask)) { *outp++ = wc; len++; }
		col++;
	  }
  }

static void DrawRowSequences(w, scol, lcol, row)
  XpwScrollTextWidget w;
  register unsigned int scol, lcol, row;
  { int cursor_col = CursorCol(w);
	if (row == CursorRow(w) && scol <= cursor_col && cursor_col < lcol)
	  {	Boolean save_cursor = _XpwTextCursorOn(w, False);
		_XpwDrawRowSequences(w, scol, lcol, row, 0);
		_XpwTextCursorOn(w, save_cursor);
	  }
	else
		_XpwDrawRowSequences(w, scol, lcol, row, 0);
  }

static void ClearOnActive(w)
  XpwScrollTextWidget w;
  { onactive_widget = NULL;
	if (onactive_mode & (ACT_INVERT|ACT_BOX))
		DrawRowSequences(w, onactive_start_col, onactive_lim_col,
											onactive_row);
	if (onactive_mode & ACT_POINTER2)
	  { Display *dpy = XtDisplay(w); Window win = XtWindow(w);
		if (w->xpwcore.pointer_shape)
			XDefineCursor(dpy, win, w->xpwcore.pointer_shape);
		else
			XUndefineCursor(dpy, win);
	  }
  }

void _XpwTextTestActiveChange(w)
  XpwScrollTextWidget w;
  {
	register int coffs, col;
	int mcol = w->xpwscrolltext.mouse.column,
		mrow = w->xpwscrolltext.mouse.row,
		num_cols, num_rows = NumRows(w);
	Boolean set_new = False;
	WinRow rp;
	XpwScrollTextWidget onact_w = onactive_widget;

	if (0 <= mrow && mrow < num_rows)
	  { num_cols = NumCols(w);
		rp = RowArray(w)[mrow];
		coffs = COFFS(rp);
		col = mcol+coffs;
		if (coffs <= col && col < num_cols && (rp[col] & (ATTR_ACTIVE<<8)))
		  { /* on active char */
			if (w == onact_w && mrow == onactive_row
			 && onactive_start_col <= mcol && mcol < onactive_lim_col)
				/* within current on-active segment */
				return;
			else
				set_new = True;
		  }
	  }

	onactive_widget = NULL;

	if (w == onact_w && onactive_row < num_rows)
	  {	/* clear current on-active segment */
		ClearOnActive(w);
		if (onactive_mode & ACT_LEAVE_NOTIFY)
		  {	w->xpwscrolltext.mouse.column = onactive_start_col;
			w->xpwscrolltext.mouse.row = onactive_row;
			XtCallCallbacks((Widget)w, XtNactiveChange, (caddr_t)NULL);
		  }
	  }

	if (set_new)
	  { /* set new on-active segment -- determine its extent */
		register int lcol = col+1;
		do col--; while (coffs <= col && (rp[col] & (ATTR_ACTIVE<<8)));
		while (lcol < num_cols && (rp[lcol] & (ATTR_ACTIVE<<8))) lcol++;
		w->xpwscrolltext.mouse.column = mcol;
		w->xpwscrolltext.mouse.row = mrow;
		onactive_mode = 0;
		XtCallCallbacks((Widget)w, XtNactiveChange, &onactive_mode);
		if (onactive_mode != 0)
		  { onactive_start_col = col = col+1-coffs;
			onactive_lim_col = lcol = lcol-coffs;
			onactive_row = mrow;
			onactive_widget = w;
			if (onactive_mode & (ACT_INVERT|ACT_BOX))
				DrawRowSequences(w, col, lcol, mrow);
			if (onactive_mode & ACT_POINTER2)
				XDefineCursor(XtDisplay(w), XtWindow(w),
										w->xpwscrolltext.pointer2_shape);
		  }
	  }
  }


static void DoBlink();

void _XpwSetBlinkTimer()
  { XpwScrollTextWidget w = blink_widget_chain;
	if (blink_timerid || !w) return;
	blink_timerid =
		XtAppAddTimeOut(XtDisplayToApplicationContext(XtDisplay(w)),
													600, DoBlink, NULL);
  }

static void DoBlink()
  { XpwScrollTextWidget w = blink_widget_chain, nextw,
						*lastp = &blink_widget_chain;
	Boolean some_blinking = False;

	blink_timerid = (XtIntervalId)NULL;

	/* toggle blink phase */
	blink_on_phase = !blink_on_phase;

	while (w)
	  { WinRow *rowarr = RowArray(w);
		unsigned row, lim_row, new_start_row = MAX_ROWS, new_lim_row = 0;
		Boolean did_cursor = False, save_cursor, cursor_blinking, cursor_blink;

		if (w->xpwscrolltext.no_blink)
		  { *lastp = w;
			w = *(lastp = &w->xpwscrolltext.next_blink_widget);
			continue;
		  }

		lim_row = min(NumRows(w), w->xpwscrolltext.blink_lim_row);
		for (row = w->xpwscrolltext.blink_start_row; row < lim_row; row++)
		  { WinRow rp = rowarr[row];
			unsigned lcol = BLINK_LCOL(rp);
			if (lcol == 0) continue;

			if (did_cursor = (row == CursorRow(w)))
				save_cursor = _XpwTextCursorOn(w, False);

			blink_row_start_col = MAX_COLS; blink_row_lim_col = 0;
			_XpwDrawRowSequences(w, BLINK_SCOL(rp), lcol, row, ATTR_BLINK<<8);
			BLINK_SCOL(rp) = blink_row_start_col;
			if ((BLINK_LCOL(rp) = blink_row_lim_col) != 0)
			  { if (new_lim_row == 0) new_start_row = row;
				new_lim_row = row+1;
			  }
		  }

		cursor_blinking = ( ((w->xpwscrolltext.cursor_winchar>>8) & ATTR_BLINK)
							&& (Flags(w) & HAVE_FOCUS) );
		cursor_blink = (cursor_blinking && !skip_cursor_blink);

		if (!did_cursor && (did_cursor = cursor_blink))
			save_cursor = _XpwTextCursorOn(w, False);

		if (did_cursor)
		  { if (cursor_blink && save_cursor && !blink_on_phase)
				/* cursor on but don't draw it */
				Flags(w) |= CURSOR_ON;
			else
				_XpwTextCursorOn(w, save_cursor);
		  }

		nextw = w->xpwscrolltext.next_blink_widget;
		w->xpwscrolltext.blink_start_row = new_start_row;
		if ((w->xpwscrolltext.blink_lim_row = new_lim_row) != 0
			|| cursor_blinking)
		  { *lastp = w;
			lastp = &w->xpwscrolltext.next_blink_widget;
			some_blinking = True;
		  }
		else
			/* no blink rows -- says no longer in the chain */
			w->xpwscrolltext.next_blink_widget = (XpwScrollTextWidget) -1;

		w = nextw;
	  }

	*lastp = NULL;
	if (some_blinking) _XpwSetBlinkTimer();
	skip_cursor_blink = False;
  }

static void EnsureBlinkWidget(w)
  XpwScrollTextWidget w;
  { if (w->xpwscrolltext.next_blink_widget == (XpwScrollTextWidget)-1)
	  { w->xpwscrolltext.next_blink_widget = blink_widget_chain;
		blink_widget_chain = w;
		if (!w->xpwscrolltext.no_blink) _XpwSetBlinkTimer();
	  }
  }

/* Called from XpwScrollTextDestroy */
void _XpwTextDestroy(w)
  XpwScrollTextWidget w;
  { XpwScrollTextWidget *thisp, *lastp = &blink_widget_chain, wc;

	/* remove if on-active widget */
	if (onactive_widget == w) onactive_widget = NULL;

	/* remove if blink widget */
	if (w->xpwscrolltext.next_blink_widget == (XpwScrollTextWidget)-1) return;
	while (wc = *lastp)
	  { thisp = &wc->xpwscrolltext.next_blink_widget;
		if (wc == w)
		  { *lastp = *thisp;
			*thisp = (XpwScrollTextWidget) -1;
			return;
		  }
		else
			lastp = thisp;
	  }
  }



/*	Routines for manipulating the representation of text on the window
 */

static winchar_copy(sp, dp, len)
register WinChar *sp, *dp;
int len;
  { register WinChar *slimp = sp+len;
	if (dp < sp)
		while (sp < slimp) *dp++ = *sp++;
	else
	  { register WinChar *dlimp = dp+len;
		while (slimp > sp) *--dlimp = *--slimp;
	  }
  }

static winchar_zero(dp, len)
register WinChar *dp;
int len;
  { register WinChar *dlimp = dp+len;
	while (dp < dlimp) *dp++ = 0;
  }

static winrow_copy(sp, dp, len)
register WinRow *sp, *dp;
int len;
  { register WinRow *slimp = sp+len;
	if (dp < sp)
		while (sp < slimp) *dp++ = *sp++;
	else
	  { register WinRow *dlimp = dp+len;
		while (slimp > sp) *--dlimp = *--slimp;
	  }
  }

static normalise_row(rp, num_cols)
register WinRow rp;
register int num_cols;
  { register int coffs = COFFS(rp), len = num_cols-coffs;
	if (coffs == 0) return;
	COFFS(rp) = 0;
	winchar_copy(rp+coffs, rp, len);
	winchar_zero(rp+len, coffs);
  }

static copy_row_chars(srp, scol, drp, dcol, nchars, num_cols)
register WinRow srp, drp;
register int scol, dcol, nchars, num_cols;
  { register int trail0, dcoffs, endfree;

	scol += COFFS(srp);
	trail0 = scol+nchars-num_cols;
	if (trail0 > 0) nchars -= trail0; else trail0 = 0;

	dcoffs = COFFS(drp);
	dcol += dcoffs;
	endfree = num_cols-(dcol+nchars);
	if (endfree < 0)
	  { normalise_row(drp, num_cols);
		dcol -= dcoffs;
		if (srp == drp) scol -= dcoffs;
	  }
	else
		trail0 = min(trail0, endfree);

	if (BLINK_LCOL(drp) != 0)
	  { BLINK_SCOL(drp) = 0; BLINK_LCOL(drp) = num_cols; }

	srp += scol;
	drp += dcol;
	winchar_copy(srp, drp, nchars);
	if (trail0 != 0) winchar_zero(drp+nchars, trail0);
  }

static clear_text(w, scol, srow, ncols, nrows)
XpwScrollTextWidget w;
register int scol, srow, ncols, nrows;
  {
	int num_cols = NumCols(w);
	WinRow *rowarr = RowArray(w), rp;
	register int col, maxlen;

	if (w == onactive_widget && srow <= onactive_row
	 && onactive_row < srow+nrows)
		ClearOnActive(w);

	while (nrows--)
	  {	rp = rowarr[srow++];
		col = COFFS(rp)+scol;
		maxlen = num_cols-col;
		ncols = min(ncols, maxlen);
		if (ncols <= 0)
			continue;
		else if (scol == 0 && ncols == maxlen)
			COFFS(rp) = num_cols;
		else
		  { rp += col;
			if (ncols == 1) *rp = 0; else winchar_zero(rp, ncols);
		  }
	  }
  }

static copy_text(w, scol, srow, ncols, nrows, c_dis, r_dis)
XpwScrollTextWidget w;
register int scol, srow, ncols, nrows;
int c_dis, r_dis;
  {
	int num_cols = NumCols(w);
	WinRow *rowarr = RowArray(w);
	register int dcol = scol+c_dis, lrow = srow+nrows;
	int drow = srow+r_dis;

	if (w == onactive_widget && drow <= onactive_row
	 && onactive_row < drow+nrows)
		ClearOnActive(w);

	if (w->xpwscrolltext.blink_lim_row != 0 && r_dis != 0)
	  { w->xpwscrolltext.blink_start_row = 0;
		w->xpwscrolltext.blink_lim_row = NumRows(w);
	  }

	if (c_dis == 0 && scol == 0 && ncols == num_cols)
	  { /* Scrolling whole rows up/down */
		register int savesrc, savedst = 0, savelim = r_dis+nrows, savelen;
		/* work out which rows need saving */
		rowarr += srow;
		if (r_dis < 0)
		  { savesrc = r_dis;
			if (savelim > 0) { savedst = savelim; savelim = 0; }
		  }
		else
			savesrc = max(r_dis, nrows);

		if ((savelen = savelim-savesrc) == 1)
		  {	WinRow rp = rowarr[savesrc];
			winrow_copy(rowarr, rowarr+r_dis, nrows);
			rowarr[savedst] = rp;
		  }
		else
		  { WinRow savearr[MAX_ROWS/2];
			winrow_copy(rowarr+savesrc, savearr, savelen);
			winrow_copy(rowarr, rowarr+r_dis, nrows);
			winrow_copy(savearr, rowarr+savedst, savelen);
		  }
	  }

	else if (r_dis == 0 && (
			   (dcol == 0 && scol+ncols == num_cols)	/* left */
			|| (scol == 0 && dcol+ncols == num_cols))	/* right */
		)
	  { /* Scrolling left/right across whole rows */
		while (srow < lrow)
		  {	WinRow rp = rowarr[srow++];
			int newoffs = COFFS(rp)-c_dis;
			if (newoffs >= 0)
			  { COFFS(rp) = min(newoffs, num_cols);
				if (BLINK_LCOL(rp) != 0)
				  { BLINK_SCOL(rp) = 0; BLINK_LCOL(rp) = num_cols; }
			  }
			else
				copy_row_chars(rp, scol, rp, dcol, ncols, num_cols);
		  }
	  }

	else
	  { /* General case */
		register int drow = srow+r_dis;
		if (r_dis < 0)
		  { /* moving rows down */
			while (srow < lrow)
				copy_row_chars(rowarr[srow++], scol, rowarr[drow++], dcol,
									ncols, num_cols);
		  }
		else
		  { /* moving rows up */
			int ldrow = drow+nrows;
			while (lrow > srow)
				copy_row_chars(rowarr[--lrow], scol, rowarr[--ldrow], dcol,
									ncols, num_cols);
		  }
	  }
  }

static void draw_text(w, string, dcol, drow, nchars)
XpwScrollTextWidget w;
register unsigned char *string;
int dcol, drow;
register int nchars;
  { int dcoffs, attrs = w->xpwscrolltext.curr_attributes,
		num_cols = NumCols(w);
	register WinRow rp = RowArray(w)[drow];

	if (w == onactive_widget && drow == onactive_row) ClearOnActive(w);

	if (attrs & ATTR_BLINK)
	  { EnsureBlinkWidget(w);
		BLINK_SCOL(rp) = min(BLINK_SCOL(rp), dcol);
		BLINK_LCOL(rp) = max(BLINK_LCOL(rp), dcol+nchars);
		w->xpwscrolltext.blink_start_row =
			min(drow, w->xpwscrolltext.blink_start_row);
		w->xpwscrolltext.blink_lim_row =
			max(drow+1, w->xpwscrolltext.blink_lim_row);
	  }

	DrawString(w, string, dcol, drow, nchars, attrs, 0);

	dcoffs = COFFS(rp);
	dcol += dcoffs;
	if (num_cols-(dcol+nchars) < 0)
	  { normalise_row(rp, num_cols);
		dcol -= dcoffs;
	  }

	attrs <<= 8;
	rp += dcol;
	while (nchars--) *rp++ = *string++ | attrs;
  }

static WinChar winchar_at(w, col, row)
XpwScrollTextWidget w;
int col, row;
  { register WinRow rp = RowArray(w)[row];
	col += COFFS(rp);
	return(col < NumCols(w) ? rp[col] : 0);
  }


/* SCROLL and associated methods */

#define NormalGC(w) w->xpwscrolltext.gc_array[0]

/* first, the most generic Scrolling routine */

static Scroll(w, col, row, ncols, nrows, c_dis, r_dis)
XpwScrollTextWidget w;
register int col, row, ncols, nrows, c_dis, r_dis;
{
	Display *dpy = XtDisplay(w);
	Window win = XtWindow(w);
	int fwidth = FontWidth(w), fheight = FontHeight(w), tmp,
		num_rows = NumRows(w), num_cols = NumCols(w), copywait = 0;
	Boolean save_cursor;

	/* two complicated to faff around trying to work out if cursor needs
	   replotting, so just do it for every scroll
	*/

	save_cursor = _XpwTextCursorOn(w, False);

	row = min(row, num_rows); row = max(row, 0);
	col = min(col, num_cols); col = max(col, 0);

	if ((tmp=row+r_dis) < 0) row -= tmp;
	if ((tmp=col+c_dis) < 0) col -= tmp;

	if (nrows == 0) nrows = num_rows - row;
	else nrows = min(nrows, num_rows - row);
	if (ncols == 0) ncols = num_cols - col;
	else ncols = min(ncols, num_cols - col);

	if (row+nrows+r_dis >= num_rows) nrows = num_rows - (row+r_dis);
	if (col+ncols+c_dis >= num_cols) ncols = num_cols - (col+c_dis);

	if (ncols != 0 && nrows != 0)
	  { /* Copy the text */
		int srcx = ColToCoord(w,col), dstx = srcx+(c_dis*fwidth),
			width = ncols*fwidth, Srow = row, Drow = row+r_dis, Nrows = nrows;
		GC gc = NormalGC(w);

		copy_text(w, col, row, ncols, nrows, c_dis, r_dis);

		/* need a separate 1-row copy to/from row 0 if not contiguous */
		do
		  { int n = (Srow == 0 || Drow == 0)
					 && Row1Offset(w) != Row0Offset(w)+fheight ? 1 : Nrows;
			XCopyArea(dpy, win, win, gc, srcx,RowToCoord(w,Srow),
									width,n*fheight, dstx,RowToCoord(w,Drow));
			copywait++;
			Nrows -= n; Srow += n; Drow += n;
		  }
		while (Nrows != 0);
	  }

	/* Clear the area made by any c_dis move, and then clear the area made
		by any r_dis move.
	*/
	if (c_dis > 0 && nrows != 0)
	  { clear_text(w, col,row, c_dis,nrows);
		_XpwClearChars(w, col,row, c_dis,nrows);
	  }
	else if (c_dis < 0 && nrows != 0)
	  { int scol = col+ncols+c_dis;
		clear_text(w, scol,row, -c_dis,nrows);
		_XpwClearChars(w, scol,row, -c_dis,nrows);
	  }

	if (r_dis > 0 && ncols != 0)
	  { clear_text(w, col,row, ncols,r_dis);
		_XpwClearChars(w, col,row, ncols,r_dis);
	  }
	else if (r_dis < 0 && ncols != 0)
	  { int srow = row+nrows+r_dis;
		clear_text(w, col,srow, ncols,-r_dis);
		_XpwClearChars(w, col,srow, ncols,-r_dis);
	  }

	_XpwTextCursorOn(w, save_cursor);
	while (copywait--) _XpwTextCopyWait(w);
}

/* associated methods */

static ScrollScreenUp(w)
XpwScrollTextWidget w;
{
	Scroll(w, 0,1, 0,0, 0, -1);
}

static ScrollScreenDown(w)
XpwScrollTextWidget w;
{
	Scroll(w, 0,0, 0,0, 0, 1);
}

static ScrollScreenLeft(w)
XpwScrollTextWidget w;
{
	Scroll(w, 1,0, 0,0, -1, 0);
}

static ScrollScreenRight(w)
XpwScrollTextWidget w;
{
	Scroll(w, 0,0, 0,0, 1, 0);
}

static ScrollLines(w, start_line, num_lines, dist)
XpwScrollTextWidget w;
int start_line, num_lines, dist;
{
	Scroll(w, 0,start_line, 0, num_lines, 0, dist);
}

static ScrollTail(w, col, row, dist)
XpwScrollTextWidget w;
int row, col, dist;
{
	Scroll(w, col, row, 0, 1,  dist, 0);
}

static ScrollTails(w, col, row, num_rows, dist)
XpwScrollTextWidget w;
int col, row, num_rows, dist;
{
	Scroll(w, col, row, 0, num_rows, dist, 0);
}

static ScrollTailLeft(w, col, row)
{
	Scroll(w, col, row, 0, 1, -1, 0);
}

static ScrollTailRight(w, col, row)
{
	Scroll(w, col, row, 0, 1, 1, 0);
}

static InsertLineAtCursor(w)
XpwScrollTextWidget w;
{
	Scroll(w, 0, CursorRow(w), 0, 0, 0, 1);
}

static InsertCharAtCursor(w)
XpwScrollTextWidget w;
{
	Scroll(w, CursorCol(w), CursorRow(w), 0, 1, 1, 0);
}

static DeleteLineAtCursor(w)
XpwScrollTextWidget w;
{
	int row = CursorRow(w);
	if (row == NumRows(w)-1)
		Clear(w, 0, row, 0, 1);
	else
		Scroll(w, 0, CursorRow(w)+1, 0, 0, 0, -1);
}
static DeleteCharAtCursor(w)
XpwScrollTextWidget w;
{
	Scroll(w, CursorCol(w)+1, CursorRow(w),0,1, -1,0);
}

/* CLEAR and associated methods */

static Clear(w, col, row, ncols, nrows)
XpwScrollTextWidget w;
int row, col, nrows, ncols;
  {
	int num_cols = NumCols(w), num_rows = NumRows(w);
	Boolean DoCursor, save_cursor;

	if (col >= num_cols || row >= num_rows) return;

	ncols = min(ncols, num_cols - col);
	nrows = min(nrows, num_rows - row);

	if (ncols == 0) ncols = num_cols - col;
	if (nrows == 0) nrows = num_rows - row;

	if (DoCursor = (col <= CursorCol(w) && row <= CursorRow(w) &&
		col+ncols >= CursorCol(w) && row + nrows >= CursorRow(w)))
		save_cursor = _XpwTextCursorOn(w, False);

	clear_text(w, col, row, ncols, nrows);
	_XpwClearChars(w, col, row, ncols, nrows);

	if (DoCursor) _XpwTextCursorOn(w, save_cursor);
  }

/* related methods */

static ClearScreen(w)
XpwScrollTextWidget w;
{	Boolean save_cursor = _XpwTextCursorClear(w);
	Clear(w, 0, 0, 0, 0);
	_XpwTextCursorOn(w, save_cursor);
}

static ClearLine(w, start_line)
XpwScrollTextWidget w;
int start_line;
{
	Clear(w,0,start_line,0,1);
}


static ClearLines(w, start_line, num_lines)
XpwScrollTextWidget w;
int start_line, num_lines;
{
	Clear(w, 0, start_line, 0, num_lines);
}

static ClearTail(w, col, row)
XpwScrollTextWidget w;
int col, row;
{
	Clear(w, col, row, 0, 1);
}

static ClearTails(w, col, row, num_lines)
XpwScrollTextWidget w;
int col, row, num_lines;
{
	Clear(w, col, row, 0, num_lines);
}

static ClearChar(w, col, row)
XpwScrollTextWidget w;
int col, row;
{
	Clear(w, col, row, 1,1);
}

static ClearLineAtCursor(w)
XpwScrollTextWidget w;
{
	Clear(w, 0, CursorRow(w), 0, 1);
}
static ClearTailAtCursor(w)
XpwScrollTextWidget w;
{
	Clear(w, CursorCol(w), CursorRow(w), 0,1);
}

static ClearCharAtCursor(w)
XpwScrollTextWidget w;
{
	Clear(w, CursorCol(w), CursorRow(w), 1,1);
}


/* Write */

/* primitive text drawing routines */

/* Strings are first pre-processed to determine how much should be painted.

   If len is < 0, the string is assumed to be a poplog string, with its length
   stored 2 longwords before the string pointer. If len is 0, the whole string
   is painted. Otherwise strlen is used to calculate the maximum length of the
   string.

   The function returns the num of chars that need to be painted.
   The new string is stored in a fixed size buffer called "processed_string".

   N.B - still needs some glorification. Should check font for char definitions
*/


#define POP_SIZE_OFFSET -2
static char *processed_string;

static int process_string(w, col, offset, len, string)
XpwScrollTextWidget w;
String string;
int col, len;
{
	int max_len, num_cols = NumCols(w);
	caddr_t *ptr = (caddr_t *)string;

	/* work out maximum size of string */
	if (len < 0)
	  { /* POPLOG STRING - length is in (ptr-2) */
		max_len = (int) ptr[POP_SIZE_OFFSET];
		len = abs(len);
	  }
	else
	  { max_len = strlen(string);
		if (len == 0) len = MAX_COLS;
	  }

	if (offset < 0) offset = 0;
	if (offset >= max_len || col >= num_cols) return(0);
	if (offset+len > max_len) len = max_len - offset;
	len = min(len, num_cols-col);
	processed_string = (char *)string + offset;
	return(len);
}


/* INSERT -writes string at row, col - after moving text in that location right
   by the correct number of chars.
*/
static XpwMethodRet Insert(w, col, row, str, len)
XpwScrollTextWidget w;
int row, col, len;
String str;
  {
	Boolean DoCursor, save_cursor;
	int num_cols = NumCols(w), ncols;

	if (row >= NumRows(w) || col >= num_cols || row < 0 || col < 0)
		return(0);

	if (DoCursor = (row == CursorRow(w)))
		save_cursor = _XpwTextCursorOn(w, False);

	len = process_string(w, col, 0, len, str);

	if ((ncols = num_cols-(col+len)) != 0)
	  { Display *dpy = XtDisplay(w);
		Window win = XtWindow(w);
		int	fwidth = FontWidth(w),
			sx = ColToCoord(w,col), sy = RowToCoord(w,row);
		copy_text(w, col, row, ncols, 1, len, 0);
		XCopyArea(dpy, win, win, NormalGC(w), sx,sy, ncols*fwidth,FontHeight(w),
								 sx+len*fwidth,sy);
	  }

	draw_text(w, processed_string, col, row, len);

	if (DoCursor) _XpwTextCursorOn(w, save_cursor);
	if (ncols != 0) _XpwTextCopyWait(w);
	return(len);
  }

static InsertAtCursor(w,string,len)
XpwScrollTextWidget w;
String string;
int len;
  {
	int row = CursorRow(w), col = CursorCol(w), num_chars;
	Boolean save_cursor = _XpwTextCursorOn(w, False);

	num_chars = Insert(w, col, row, string, len);
	_XpwTextMoveCursorTo(w, col+num_chars, row);
	_XpwTextCursorOn(w, save_cursor);
  }

/* WRITE - most general case text overwrite procedure */

/* takes an array of strings. For each string, it draws the substring
   starting at offset, of length len, at row,col. If clear != 0, the line is
   cleared first.
*/

static Write(w, col, row, start_line, nlines, strings, offset, len, clear)
XpwScrollTextWidget w;
int col,row,offset,len,start_line, nlines,clear;
String strings[];
  {
	Boolean DoCursor = False, save_cursor;
	int i, cursor_row = CursorRow(w),
		num_cols = NumCols(w), num_rows = NumRows(w);

	if (row >= num_rows || col >= num_cols || row < 0 || col < 0) return;
	if (row + nlines > num_rows) nlines = num_rows - row;
	if (len + col > num_cols) len = num_cols - col;

	if (DoCursor = (cursor_row >= row && cursor_row <= (row+nlines)))
		save_cursor = _XpwTextCursorOn(w, False);

	if (clear) Clear(w, col, row, num_cols-col, nlines);

	for (i=start_line; i < start_line+nlines; i++)
	  { int l = process_string(w, col, offset, len, (char *)strings[i]);
		draw_text(w, processed_string, col, row, l);
		row++;
	  }

	if (DoCursor) _XpwTextCursorOn(w, save_cursor);
  }


/* writes str at row,col. If clear != 0, the whole line is cleared first.
*/

static WriteLine(w, col, row, str, len, clear)
XpwScrollTextWidget w;
int row, col, len, clear;
char *str;
  {
	String strings[1];
	strings[0] = str;
	Write(w,col,row,0,1,strings,0,len,clear);
  }

/* same as above, for multiple lines */

static WriteLines(w, col, row, start_line,num_lines, strs, len, clear)
XpwScrollTextWidget w;
int row, col, clear,len,start_line, num_lines;
String strs[];
  {
	Write(w,col,row,start_line, num_lines,strs,0,len,clear);
  }

/* takes a string, and writes substring starting at offset, of length len, at
   row,col.
*/

static WriteSubstr(w, col, row, string, offset, len, clear)
XpwScrollTextWidget w;
int col,row,offset,len,clear;
String string;
  {	String strings[1];
	strings[0] = string;
	Write(w,col,row,0,1,strings,offset,len,clear);
  }

	/* really a variant of write, but hard coded for efficiency */
static WriteAtCursor(w, str, len)
XpwScrollTextWidget w;
String str;
int len;
  { int col = CursorCol(w), row = CursorRow(w), num_cols = NumCols(w);
	Boolean save_cursor;
	if (row >= NumRows(w) || col >= num_cols || row < 0 || col < 0)
		return;

	save_cursor = _XpwTextCursorOn(w, False);

	len = process_string(w, col, 0, len, str);
	draw_text(w, processed_string, col, row, len);

	col += len;
	CursorCol(w) = min(col, num_cols-1);
	_XpwTextCursorOn(w, save_cursor);
 }


/* CURSOR CONTROL */

Boolean _XpwTextCursorOn(w, on)
  XpwScrollTextWidget w;
  Boolean on;
  { unsigned flags = Flags(w);
	Boolean wason = flags & CURSOR_ON, hasfocus = flags & HAVE_FOCUS;
	WinChar wc, cwc = w->xpwscrolltext.cursor_winchar;
	unsigned char cc = cwc & 0xff, tmp[1];
	int cattrs = cwc >> 8, col = CursorCol(w), row = CursorRow(w);

	if (hasfocus && (cattrs & ATTR_BLINK))
	  { EnsureBlinkWidget(w);
		if (on) skip_cursor_blink = True;
	  }

	if (wason)
	  {	/* cursor is already on */
		if (!on)
		  { /* turn it off */
			if ((wc = winchar_at(w, col, row)) != 0)
			  { tmp[0] = wc;
				DrawString(w, tmp, col, row, 1, wc>>8, 0);
			  }
			else
				_XpwClearChars(w, col, row, 1, 1);
		  }
	  }

	else
		/* cursor is currently off */
		if (on)
		  { /* turn it on */
			int dsflags = DS_ISCURSOR|DS_CURSORCHAR|DS_NOBACK,
				cgc_index = cattrs & ATTR_GC_MASK;

			if (hasfocus)
			  { cattrs &= ~cgc_index;
				cgc_index = GCIndex(cgc_index/NFONTS,
									(cgc_index%NFONTS)|BOLDFONTBIT);
				cattrs |= cgc_index;
			  }

			if (cc == 'O' && ((cgc_index%NFONTS) & BOLDFONTBIT)
				&& (wc = winchar_at(w, col, row)) != 0
				&& (!((wc>>8) & ATTR_BLINK)
					|| w->xpwscrolltext.no_blink || blink_on_phase) )
			  { /* Draw the underlying char in inverse */
				int attrs = wc>>8, gc_index = attrs & ATTR_GC_MASK;
				cattrs = (attrs & ~gc_index)
							| GCIndex(cgc_index/NFONTS, gc_index%NFONTS);
				dsflags = DS_ISCURSOR|DS_INVERT;
				cc = wc;
			  }

			tmp[0] = cc;
			DrawString(w, tmp, col, row, 1, cattrs & ~ATTR_BLINK, dsflags);
		  }

	Flags(w) = (on ? flags | CURSOR_ON : flags & ~CURSOR_ON);
	return(wason);
  }

Boolean _XpwTextCursorClear(w)
XpwScrollTextWidget w;
  { Boolean wason = Flags(w) & CURSOR_ON;
	if (wason)
	  { _XpwClearChars(w, CursorCol(w), CursorRow(w), 1, 1);
		Flags(w) &= ~CURSOR_ON;
	  }
	return(wason);
  }


XpwMethodRet _XpwTextMoveCursorTo(w, col, row)
XpwScrollTextWidget w;
int col, row;
{
	int num_cols = NumCols(w), num_rows = NumRows(w);
	Boolean save_cursor;

	col = min(num_cols-1,col);
	row = min(num_rows-1,row);
	col = max(0, col);
	row = max(0,row);

	if (row == CursorRow(w) && col == CursorCol(w)) return(0);

	save_cursor = _XpwTextCursorOn(w, False);
	CursorCol(w) = col;
	CursorRow(w) = row;
	_XpwTextCursorOn(w, save_cursor);
	return(0);
}

static MoveCursorUp(w)
XpwScrollTextWidget w;
  { _XpwTextMoveCursorTo(w, CursorCol(w), CursorRow(w) -1); }

static MoveCursorDown(w)
XpwScrollTextWidget w;
  { _XpwTextMoveCursorTo(w, CursorCol(w), CursorRow(w) +1); }

static MoveCursorLeft(w)
XpwScrollTextWidget w;
  { _XpwTextMoveCursorTo(w, CursorCol(w)-1, CursorRow(w)); }

static MoveCursorRight(w)
XpwScrollTextWidget w;
  { _XpwTextMoveCursorTo(w, CursorCol(w)+1, CursorRow(w)); }


/* FONT CHANGES */

static SetFont(w, string, resize, redisplay)
XpwScrollTextWidget w;
char *string;
int resize, redisplay;
  { /* might resize widget to make it fit nicely */
  }

static Bell(w,volume)
Widget w;
int volume;
  { XBell(XtDisplay(w), volume); }


/*******************************************************************
 * Text Selections
 *******************************************************************/

static XpwAssocTable *setsel_table = NULL;

typedef struct setsel_context {
	struct setsel_context *next_widget;
	Widget			widget;
	String			string;
	int				length;
	Time			time;
	XtCallbackProc	lose_p;
	Opaque			client;
} setsel_context;

static setsel_context *GetSelnContext(w, selection, create)
Widget w;
XpwAssocID selection;
Boolean create;
  {	setsel_context *c, *context;

	/* build table first time */
	if (!setsel_table) setsel_table = XpwCreateAssocTable(4);

	context = c = (setsel_context*)XpwLookupAssoc(setsel_table, selection);

	while (c != NULL)
		if (c->widget == w || c->widget == NULL)
		  { if (c->next_widget != NULL)
			  { /* there can only be one more */
				XtFree(c->next_widget->string);
				XtFree((char *)c->next_widget);
				c->next_widget = NULL;
			  }
			return((c->widget == w || create) ? c : NULL);
		  }
		else
			c = c->next_widget;

	/* build new context for selection */
	if (create)
	  { c = XtNew(setsel_context);
		c->next_widget = context;
		c->widget	= (Widget)NULL;
		c->string	= (String)NULL;
		c->length	= 0;
		c->time		= 0;
		c->lose_p	= (XtCallbackProc)NULL;
		XpwMakeAssoc(setsel_table, (XpwAssocID)selection, (caddr_t)c);
	  }
	return(c);
}

static Boolean SelectionConvert(w, selectionp, target, type_return,
								value_return, length_return, format_return)
Widget w;
XpwAssocID *selectionp;
Atom *target;
Atom *type_return;
Opaque *value_return;
unsigned long *length_return;
int *format_return;
  {
	Display *dpy = XtDisplay(w);
	Atom targets = XInternAtom(dpy, "TARGETS", False);
	Boolean result = False;
	setsel_context *context = GetSelnContext(w, *selectionp, False);

	if (context == NULL)
		return(False);
	else if (*target == XA_STRING ||
			*target == XInternAtom(dpy,"TEXT",False))
	  { /* Return the string */
		*value_return = (Opaque)XtNewString(context->string);
		*length_return = context->length;
		*type_return = XA_STRING;
		*format_return = 8;
		result = True;
	  }
	else if (*target == XInternAtom(dpy, "LENGTH", False))
	  { /* Return the length of the string */
		*value_return = (Opaque)&context->length;
		*type_return = XA_INTEGER;
		*length_return = 1;
		*format_return = sizeof(int) * 8;
		result = True;
	  }
	else if (*target == XInternAtom(dpy, "LIST_LENGTH", False))
	  { /* Return number of elements, = 1 */
		static int len = 1;
		*value_return = (Opaque)&len;
		*type_return = XA_INTEGER;
		*length_return = 1;
		*format_return = sizeof(int) * 8;
		result = True;
	  }
	else if (*target == XInternAtom(dpy, "TIMESTAMP", False))
	  { /* Return time of selection */
		static int len = 1;
		*value_return = (Opaque)&context->time;
		*type_return = XA_INTEGER;
		*length_return = 1;
		*format_return = sizeof(int) * 8;
		result = True;
	  }
	else if (*target == targets)
	  { /* Return the list of things I can return */
		Atom *atomPtr;
		*value_return = (Opaque)(XtMalloc(sizeof(Atom)*7));
		*length_return = 7;
		atomPtr = *(Atom**)value_return;
		*atomPtr++ = XA_STRING;
		*atomPtr++ = targets;
		*atomPtr++ = XInternAtom(dpy, "TEXT", False);
		*atomPtr++ = XInternAtom(dpy, "MULTIPLE", False);
		*atomPtr++ = XInternAtom(dpy, "LENGTH", False);
		*atomPtr++ = XInternAtom(dpy, "LIST_LENGTH", False);
		*atomPtr++ = XInternAtom(dpy, "TIMESTAMP", False);
		*type_return = XA_ATOM;
		*format_return = sizeof(Atom) * 8;
		result = True;
	  }

	return(result);
  }

/* Losing selections */
static void LoseSelection(w, selectionp)
Widget w;
XpwAssocID *selectionp;
{ 	setsel_context *context = GetSelnContext(w, *selectionp, False);
	XtCallbackProc lose_p = context->lose_p;
	XtFree(context->string);
	context->widget = (Widget)NULL;
	context->string = (String)NULL;
	context->length = 0;
	context->time	= 0;
	context->lose_p	= (XtCallbackProc)NULL;
	if (lose_p != NULL)
		(*lose_p)(w, context->client, selectionp);
}

XpwMethodRet XpwSetSelection(w, location, string, nbytes, lose_p, client)
Widget w;
String location;
String string;
int nbytes;
XtCallbackProc lose_p;
Opaque client;
  {
	Display *dpy = XtDisplay(w);
	Atom selection;
	setsel_context *context;

	/* convert selection location - 1 and 2 are left alone, but rest
	   are treated as strings
	*/
	selection = (Atom)location;
	if (!selection)
		selection = XA_PRIMARY;
	else if (selection > 2)
		selection = XInternAtom(dpy, location, FALSE);


	if (string && nbytes)
	  { /* specified a new string to set the selection as */
		context = GetSelnContext(w, (XpwAssocID)selection, True);
		XtFree(context->string);	/* free any old string */
		/* copy string, and set length */
		context->string = XtNewString(string);
		context->length = nbytes;
		context->time = XtLastTimestampProcessed(dpy);
		context->lose_p	= lose_p;
		context->client	= client;

		if (context->widget != NULL) return; /* already own the selection */

		/* n.b. the selection data must be set up before calling
		 * XtOwnSelection, because xclipboard does nasty things like
		 * having its lose procedure get the new selection and then
		 * immediately taking back ownership (so that our call to
		 * XtOwnSelection returns true even though we've lost it again!)
		 */
		context->widget = w;
		if (!XtOwnSelection(w, selection, context->time,
							SelectionConvert, LoseSelection, NULL))
			/* failed */
			LoseSelection(w, &selection);
	  }

	else
		/* clearing selection */
	  { context = GetSelnContext(w, (XpwAssocID)selection, False);
		if (context != NULL)
		  { XtDisownSelection(w, selection, context->time);
			/* just to make sure */
			XtFree(context->string); /* free old string */
			context->widget = (Widget)NULL;
			context->string = (String)NULL;
			context->length = 0;
			context->time	= 0;
			context->lose_p	= (XtCallbackProc)NULL;
		  }
	  }

}


/* Receiving selections */
typedef struct {XtCallbackProc p; Opaque val;} closure;

static void SelectionCB(w, clos, selectionp, typep, value, length, format)
Widget w;
closure *clos;
Atom *selectionp;
Atom *typep;
Opaque value;
unsigned long *length;
int *format;
  { if (*typep != XA_STRING) value = NULL;
	(*clos->p)(w, clos->val, &value);
	XtFree(value);
	XtFree((char *)clos);
  }

XpwMethodRet XpwGetSelection(w, location, receive_p, client)
Widget w;
String location;
XtCallbackProc receive_p;
Opaque client;
  { Display *dpy = XtDisplay(w);
	Time time = XtLastTimestampProcessed(XtDisplay(w));
	Atom selection = (Atom)location;
	setsel_context *context;

	if (!selection)
		selection = XA_PRIMARY;
	else if (selection > 2)
		selection = XInternAtom(dpy, location, FALSE);

	context = GetSelnContext(w, (XpwAssocID)selection, False);

	if (context != NULL)
		/* call this directly */
		(*receive_p)(w, client, &context->string);
	else
	  { closure *clos = XtNew(closure);
		clos->p = receive_p; clos->val = client;
		XtGetSelectionValue(w, selection, XA_STRING, SelectionCB, clos, time);
	  }
	return(0);		/* dummy */
  }


/****************************************************************
 *
 * Methods
 *
 ****************************************************************/

externaldef (xpwscrolltextmethods)
XpwMethod _xpwScrollTextMethods[] = {
/*	{id, 				proc, 		 	num_args, 	flags	 }*/
#define M METHOD_STRUCT
M	(XpwMScroll,		Scroll,					6, RequiresWidget|Cut),
M	(XpwMScrollScreenUp,	ScrollScreenUp,		0, RequiresWidget|Cut),
M	(XpwMScrollScreenDown,	ScrollScreenDown,	0, RequiresWidget|Cut),
M	(XpwMScrollScreenLeft,	ScrollScreenLeft,	0, RequiresWidget|Cut),
M	(XpwMScrollScreenRight,	ScrollScreenRight,	0, RequiresWidget|Cut),
M	(XpwMScrollLines,	ScrollLines,		3, RequiresWidget|Cut),
M	(XpwMScrollTail,	ScrollTail,		3, RequiresWidget|Cut),
M	(XpwMScrollTails,	ScrollTails,		4, RequiresWidget|Cut),
M	(XpwMScrollTailLeft,	ScrollTailLeft,		2, RequiresWidget|Cut),
M	(XpwMScrollTailRight,	ScrollTailRight,	2, RequiresWidget|Cut),
M	(XpwMInsertLineAtCursor,InsertLineAtCursor,	0, RequiresWidget|Cut),
M	(XpwMInsertCharAtCursor,InsertCharAtCursor,	0, RequiresWidget|Cut),
M	(XpwMDeleteLineAtCursor,DeleteLineAtCursor,	0, RequiresWidget|Cut),
M	(XpwMDeleteCharAtCursor,DeleteCharAtCursor,	0, RequiresWidget|Cut),
M	(XpwMClear,			Clear,			4, RequiresWidget|Cut),
M	(XpwMClearScreen,	ClearScreen,		0, RequiresWidget|Cut),
M	(XpwMClearLine,		ClearLine,		1, RequiresWidget|Cut),
M	(XpwMClearLines,	ClearLines,		2, RequiresWidget|Cut),
M	(XpwMClearTail,		ClearTail,		2, RequiresWidget|Cut),
M	(XpwMClearTails,	ClearTails,		3, RequiresWidget|Cut),
M	(XpwMClearChar,		ClearChar,		2, RequiresWidget|Cut),
M	(XpwMClearLineAtCursor,	ClearLineAtCursor,	0, RequiresWidget|Cut),
M	(XpwMClearTailAtCursor,	ClearTailAtCursor,	0, RequiresWidget|Cut),
M	(XpwMClearCharAtCursor,	ClearCharAtCursor,	0, RequiresWidget|Cut),
M	(XpwMInsert,		Insert,			4, RequiresWidget|Cut),
M	(XpwMInsertAtCursor,	InsertAtCursor,		2, RequiresWidget|Cut),
M	(XpwMWrite,			Write,			8, RequiresWidget|Cut),
M	(XpwMWriteLine,		WriteLine,		5, RequiresWidget|Cut),
M	(XpwMWriteLines,	WriteLines,		7, RequiresWidget|Cut),
M	(XpwMWriteSubstr,	WriteSubstr,		6, RequiresWidget|Cut),
M	(XpwMWriteAtCursor,	WriteAtCursor,		2, RequiresWidget|Cut),
M	(XpwMCursorTo,		_XpwTextMoveCursorTo,	2, RequiresWidget|Cut),
M	(XpwMCursorUp,		MoveCursorUp,		0, RequiresWidget|Cut),
M	(XpwMCursorDown,	MoveCursorDown,		0, RequiresWidget|Cut),
M	(XpwMCursorLeft,	MoveCursorLeft,		0, RequiresWidget|Cut),
M	(XpwMCursorRight,	MoveCursorRight,	0, RequiresWidget|Cut),
M	(XpwMBell,				Bell,					1, RequiresWidget|Cut),
M	(XpwMSetCharAttributes,	_XpwSetCharAttributes,	1, RequiresWidget|Cut),
M	(XpwMGetCharAttributes,	_XpwGetCharAttributes,	0, RequiresWidget|Cut),
M	(XpwMSetTextCursor,		_XpwSetTextCursor,		1, RequiresWidget|Cut),
M	(XpwMGetTextCursor,		_XpwGetTextCursor,		0, RequiresWidget|Cut),
M	(XpwMSetTextSelection,	XpwSetSelection,		3, RequiresWidget),
M	(XpwMGetTextSelection,	XpwGetSelection,		1, RequiresWidget),
#undef M
};

externaldef (_num_xpwscrolltextmethods)
int _num_xpwScrollTextMethods = XtNumber(_xpwScrollTextMethods);



/* --- Revision History ---------------------------------------------------
--- John Gibson, Oct  9 1995
		Added active attribute stuff and drawing shadowed chars etc.
		Also added fixed-width drawing of proportional fonts or fonts
		less than FontWidth in width using XDrawText.
--- John Gibson, Jan  9 1994
		Added support for different cursors/blinking cursor etc.
--- John Gibson, Dec 19 1993
		Fixed ANSI warnings
--- Julian Clinton, Mar 29 1993
		Modified _XClearRectangle so that it doesn't grope into the internals
		of a GC.
--- John Gibson, Jul 28 1992
		Changed Scroll so that it deals sensibly with the case where
		row = num_rows and nrows = 0, although r_dis is non-zero
		(i.e. there is no text to copy, but just an area to clear).
--- John Gibson, Mar  9 1992
		VMS externaldef mods
--- John Gibson, Feb 12 1992
		Added drawing of diamond in DrawGraphString
--- John Gibson, Feb  2 1992
		Changes to support 8 per-char colours
--- John Gibson, Jan 25 1992
		Added blinking attribute
--- John Gibson, Jan 15 1992
		Added support for mark fg & bg colours in col 0
--- John Gibson, Jan 13 1992
		Add alt font support
--- John Gibson, Dec 24 1991
		Added support for built-in graphics chars
--- John Gibson, Dec 18 1991
		Added DrawString to support bold & underline modes
--- Simon Nichols, Nov 27 1991
		Changed semicolon to comma in middle of declaration starting at
		line 38.
--- John Gibson, Nov 21 1991
		Moved _xpwScrollTextMethods and _num_xpwScrollTextMethods to
		end of file
--- John Gibson, Oct 31 1991
		VMS mods
--- John Gibson, Oct 23 1991
		Changed declaration of -string- in -draw_text- to unsigned char
--- John Gibson, Sep  4 1991
		Fixed bug in Insert where _XpwTextCopyWait was being called when
		no XCopyArea had been done (causing it to hang up).
--- John Gibson, Aug 21 1991
		Rewrote selections (again)
--- John Gibson, Aug 14 1991
		ClearScreen now indexed on XpwMClearScreen not XpwMClearWindow
--- John Gibson, Aug  5 1991
		Cursor highlighting now dependent on flags field
--- John Gibson, Jul 27 1991
		Now represents text window contents in row array in widget.
		Large number of changes for this and to fix other bugs.
--- John Gibson, Jul 16 1991
		Rewrote selection code
--- John Gibson, Jul 14 1991
		Changes to cursor state handling
--- John Gibson, Jul 11 1991
		Added initialisation of context fields in RestoreSelectionVars
		(it helps not to call XtFree on any old junk).
--- Jonathan Meyer, Jul  9 1991
	Fixed call to LoseSelection inside SetSelection
--- Jonathan Meyer, Jul  8 1991
	Added setsel_context to allow multiple selections to be maintained
	at once
--- Jonathan Meyer, Jul  3 1991
	Called LoseSelection if SetSelection fails. Made SetSelection
	with a NULL string disown the selection if necessary
--- Jonathan Meyer, Jul  2 1991
	Wrote _XpwTextSetColors
--- Jonathan Meyer, Jun 27 1991
	Added set_geom arg to _XpwTextRecalc
--- Jonathan Meyer, Jun 17 1991
	Renamed Xpt* Xpw
--- Jonathan Meyer, Jun  4 1991
	Added SetColors, now sensitive to cursor position
--- Jonathan Meyer, Aug 21 1990
	Renamed Text.c, added "X11" prefix to #include's.
	Added XpwNotInstalled.
	XpwTextRecalc -> _XpwTextRecalc,
	XpwTextCursor -> _XpwTextCursor,
--- Andreas Schoter, Jul 23 1990
	Changed all pop* identifiers to xpw*
--- Ian Rogers, Jul 20 1990 - changed Pop* to Xpw*
 */
