/* --- Copyright University of Sussex 1996. All rights reserved. ----------
 * File:            C.x/x/Xpw/XpwScrTextP.h
 * Version:         Revision 5.0
 * Purpose: 	    Private header file for the XpwScrollText widget.
 * Author:          Jonathan Meyer (see revisions)
 * Date:            11 February 1989
 * Documentation:   REF *XpwScrText, HELP *XpwScrText
 * Related Files:   XpwScrText.h XpwScrText.c xpwgraphics.p
 */

/*
 * Notes:
 *
 *  This contains all of the private declarations.
 *  It is organised as follows:
 *  	o  new fields for instances of the XpwComposite - a structure.
 *  	o  declaration of the full instance record
 *  	o  new fields for the class XpwComposite - a structure.
 *  	o  declaration of the full class record.
 *	o  private macro definitions used in widget source.
 */

#ifndef _XpwScrollTextP_h
#define _XpwScrollTextP_h

#ifdef XpwNotInstalled
#include "XpwScrText.h"
#include "XpwCoreP.h"
#else
#include <X11/Xpw/XpwScrText.h>
#include <X11/Xpw/XpwCoreP.h>
#endif

/* Window rows */
typedef unsigned short WinChar;
typedef WinChar *WinRow;
#define COFFS(rp)	(rp)[-1]
#define BLINK_SCOL(rp)	(rp)[-2]
#define BLINK_LCOL(rp)	(rp)[-3]

/* Character GC index */
#define BOLDFONTBIT       1
#define NORMAL_FONT       (0<<1)
#define ALT_FONT          (1<<1)
#define BOLD_NORMAL_FONT  (NORMAL_FONT|BOLDFONTBIT)
#define BOLD_ALT_FONT     (ALT_FONT|BOLDFONTBIT)
#define NFONTS 4

/* Normal colours range from 0 - NCOLORS-1 (normal colour is 0) */
#define NCOLORS 8
/* ... then colours 8 - 15 represent colours with the active bit set */
#define NALLCOLORS	16
#define CURSOR_COLORNUM NALLCOLORS

#define GCIndex(colornum, fontnum)  ((colornum)*NFONTS+(fontnum))

/* Allow shared gcs for 3 colours in all fonts for main text, and 1 colour
 * in all fonts for status text
 */
#define NNORMALGCS      GCIndex(3, 0)
#define NSTATUSGCS      GCIndex(1, 0)
#define CURSOR_GCINDEX  GCIndex(CURSOR_COLORNUM, 0)

/* Character GC index + other attributes */
#define ATTR_GC_MASK   0x1f       /* 5 bits to hold < NCOLORS*NFONTS */
#define ATTR_ACTIVE    (1<<5)
#define ATTR_UNDERLINE (1<<6)
#define ATTR_BLINK     (1<<7)

#define MAX_ROWS       512
#define MAX_COLS       1024


/* flags field */
#define CURSOR_ON    (1<<0)   /* Cursor is currently displayed */
#define ACT_FOCUS    (1<<1)   /* we are the actual focus window */
#define PTR_FOCUS    (1<<2)   /* we inherit the focus by having the pointer */
#define SYN_FOCUS    (1<<3)   /* synthetic focus courtesy of XtSetKeyboardFocus */
#define CALLBACK_YES (1<<4)   /* Last callback said client had focus */
#define HAVE_FOCUS (ACT_FOCUS | PTR_FOCUS | SYN_FOCUS)
#define MONOCHROME   (1<<5)   /* Monochrome screen */

typedef struct { Pixel fg, bg; } ColorPair;

typedef struct {
	unsigned short colornum;
	Boolean underline, bold, alt_font, blink, active;
} AttrResources;

typedef struct {
	int x, y, column, row;
} MousePosition;

/* this is directly below the XpwCore in widget heirarchy */

/* New fields for the poplog scrolltext widget instance record */

typedef struct {
	GC cursor_gc, status_cursor_gc;
	GC gc_array[NNORMALGCS];
	GC status_gc_array[NSTATUSGCS];
	GC work_gc;
	short h_margin, v_margin;
	Position col0_offset, row0_offset, row1_offset, split_offset;
	int font_width, font_height, font_baseline;
	int num_columns, num_rows;
	int cursor_column, cursor_row;
	MousePosition mouse;
	unsigned char flags;
	Boolean cursor_status;
	Boolean is_fully_visible;
	Boolean synthetic;
	Boolean draw_graphic_chars;
	unsigned short font_draw_modes;
	String key;
	XtCallbackList keyboard_event;
	XtCallbackList button_event;
	XtCallbackList motion_event;
	Pixel status_foreground, status_background, cursor_color;
	ColorPair alt_colors[NALLCOLORS-1];
	unsigned short curr_attributes;
	AttrResources curr_attr_resources;
	Boolean highlight_on;
	Boolean round_size;
	Boolean no_blink;
	Boolean no_gray_scale;
	unsigned short blink_start_row, blink_lim_row;
	XpwScrollTextWidget next_blink_widget;
	WinRow *row_array;
	XtCallbackList focus_change;
	String geometry;
	XFontStruct *alt_fonts[3];

	/* Blank resources - these resources are dummies filled by XVed */
	Boolean menubar_on, scrollbar_on, hscrollbar_on;
	String auto_geometry;
	Pixel scrollbar_foreground, scrollbar_background;
	Pixel menubar_foreground, menubar_background;

	unsigned short selection_color_num, status_style;
	WinChar cursor_winchar;
	unsigned int cursor_char;
	unsigned int default_color_mask;
	XtCallbackList active_change;
	Cursor pointer2_shape;
	Pixel shadow_whites[2];
	unsigned short shadow_white_bits[2];
	unsigned int draw_shadow_mask;
} XpwScrollTextPart;

/* Full instance record declaration */
typedef struct _XpwScrollTextRec {
	CorePart core;
	XpwCorePart xpwcore;
	XpwScrollTextPart xpwscrolltext;
} XpwScrollTextRec;

/* New fields for the class record - there are none.*/
typedef struct {int dummy;} XpwScrollTextClassPart;

/* Full class record declaration. */
typedef struct _XpwScrollTextClassRec {
	CoreClassPart core_class;
	XpwCoreClassPart xpwcore_class;
	XpwScrollTextClassPart xpwscrolltext_class;
} XpwScrollTextClassRec;

/* Class pointer. */
externalref XpwScrollTextClassRec xpwScrollTextClassRec;

/* Useful MACROS */

#define Col0Offset(w)   ((w)->xpwscrolltext.col0_offset)
#define Row0Offset(w)   ((w)->xpwscrolltext.row0_offset)
#define Row1Offset(w)   ((w)->xpwscrolltext.row1_offset)
#define SplitOffset(w)  ((w)->xpwscrolltext.split_offset)
#define NumRows(w)      ((w)->xpwscrolltext.num_rows)
#define NumCols(w)      ((w)->xpwscrolltext.num_columns)
#define CursorRow(w)    ((w)->xpwscrolltext.cursor_row)
#define CursorCol(w)    ((w)->xpwscrolltext.cursor_column)
#define FontWidth(w)    ((w)->xpwscrolltext.font_width)
#define FontHeight(w)   ((w)->xpwscrolltext.font_height)
#define FontBaseline(w) ((w)->xpwscrolltext.font_baseline)
#define RowArray(w)     ((w)->xpwscrolltext.row_array)
#define Flags(w)        ((w)->xpwscrolltext.flags)
#define StatusStyle(w)  ((w)->xpwscrolltext.status_style)

#define ColToCoord(w,col) (FontWidth(w)*(col) + Col0Offset(w))

#define MainRowToCoord(w,row) (FontHeight(w)*((row)-1) + Row1Offset(w))
#define RowToCoord(w,row)    ((row)==0 ? Row0Offset(w) : MainRowToCoord(w,row))

#define FontStruct(w,fontnum) ((fontnum)==NORMAL_FONT ? w->xpwcore.font \
				: w->xpwscrolltext.alt_fonts[(fontnum)-1])

#define FontFieldPtr(w,fontnum) ((fontnum)==NORMAL_FONT ? &w->xpwcore.font \
				: &w->xpwscrolltext.alt_fonts[(fontnum)-1])

Boolean _XpwTextCursorOn(), _XpwTextCursorClear();

#define max(a, b) ((int)(a) > (int)(b) ? (a) : (b))
#define min(a, b) ((int)(a) < (int)(b) ? (a) : (b))



#endif /* _XpwScrollTextP_h */



/* --- Revision History ---------------------------------------------------
--- John Gibson, Apr 16 1996
	Added auto_geometry field
--- John Gibson, Oct  9 1995
	Various changes.
--- John Gibson, Feb 11 1994
	Added hscrollbar_on field
--- John Gibson, Jan  9 1994
	Changes for multiple cursors, different status line positions etc
--- John Gibson, Dec 19 1993
	Cast min & max params as (int)
--- John Gibson, Jun 23 1992
	Added no_gray_scale field
--- John Gibson, Feb  2 1992
	Changes to support 8 per-char colours
--- John Gibson, Jan 27 1992
	Added blink stuff
--- John Gibson, Jan 13 1992
	Added alt font stuff and regions
--- John Gibson, Dec 23 1991
	Added draw_graphic_chars field
--- John Gibson, Dec 18 1991
	Changes to support bold & underline modes
--- John Gibson, Nov  4 1991
	Added geometry field
--- John Gibson, Oct 31 1991
	extern -> externalref for XpwScrollTextClassRec
--- John Gibson, Oct 24 1991
	Removed casts on field macros (not needed)
--- John Gibson, Sep 29 1991
	Added round_size
--- Jonathan Meyer, Sep 11 1991 Added menubar_foreground/background
--- John Gibson, Aug  6 1991
	Added h_margin and v_margin as resource margin fields -- col0_offset
	and rowX_offset are actual margins used.
--- John Gibson, Aug  6 1991
	Added focus_change field
--- John Gibson, Aug  5 1991
	Replaced cursor_state field with flags
--- Jonathan Meyer, Aug  2 1991
	Added scrollbar_foreground, scrollbar_background
--- John Gibson, Jul 27 1991
	Added row_array and new GC fields to widget, removed exposed_lines,
	etc.
--- John Gibson, Jul 14 1991
	Changes to cursor state handling
--- Jonathan Meyer, Jul  8 1991
	Added scrollbar_on, menubar_on (dummy entries used by XVed)
--- Jonathan Meyer, Jun  4 1991
	Added synthetic
--- Jonathan Meyer, Aug 21 1990
	Added test for XpwNotInstalled.
--- Andreas Schoter, Jul 23 1990
	Changed all pop* identifiers to xpw*
--- Andreas Schoter 16 July 1990
	Replaced all occurances of Pop* variable names with Xpw*
*/
