/* --- Copyright University of Sussex 1995. All rights reserved. ----------
 > File:            C.x/x/pop/auto/XpmImportString.p
 > Purpose:         Support for Motif strings
 > Author:          Jonathan Meyer, Feb  8 1991 (see revisions)
 > Documentation:   HELP * MOTIF
 */
compile_mode :pop11 +strict;

section;
exload_batch;

include xpt_coretypes.ph;
uses fast_xt_util;


;;; These are duplicated from XmConstants.ph -- NOTE that this file can't
;;; include that, since it uses this file.

l_typespec		;;; duplicated from XmConstants.ph
	XmStringDirection: byte,
	XmStringCharSet :XptString;

lconstant macro XmSTRING_DEFAULT_CHARSET = nullstring;


/************************************************************************
 *  Support for Motif Strings
 ************************************************************************/

;;; coerce external pointer into motif string
define XpmImportString =
	XptImportAny(%"XmString"%);
enddefine;

XptPopLoadProcedures XpmImportString [^^XM_EXLIBS]
	XmStringCreateLtoR(x,y) :exptr#XpmImportString,
	XmStringInitContext(x,y) :XptBoolean,
	XmStringGetNextSegment(v,w,x,y,z) :XptBoolean,
	XmStringFreeContext(x) :void,
	XmStringFree(x) :void,
;

define XpmCheckString =
	XptLiveTypeCheck(%"XmString"%)
enddefine;

;;; coerce motif string -> popstring
;;; in order to be re-entrant, we don't use lconstant ptrs, but get fresh
;;; ones for each call. We cache those pointers in a dynamic list, so
;;; as to not generate too much garbage.
lvars fixed_ptrs = pdtolist(exptr_copy_fixed(%null_external_ptr%));

define XpmCoerceString(xmstring) -> string;
	lvars xmstring, string, i, c, context, text;

	lconstant
		;;; we never get these parameters, so lconstant ones are ok.
		charset_ptr 	= EXPTRINITSTR(:XmStringCharSet),
		direction_ptr 	= EXPTRINITSTR(:XmStringDirection),
	;

	;;; get three pointers from the free list
	lvars (context_ptr, text_ptr, separator_ptr)
			= (repeat 3 times dest(fixed_ptrs) -> fixed_ptrs endrepeat);

	external_ptr_props(context_ptr) -> context;
	external_ptr_props(text_ptr) -> text;

	if xmstring.isstring then
		xmstring -> string;
	elseif XmStringInitContext(context_ptr, xmstring) then
		consstring(#|
			while XmStringGetNextSegment(context, text_ptr,
								charset_ptr, direction_ptr, separator_ptr) do
				1 -> i;
				until (exacc :byte[] text[i] ->> c) == 0 do
					i fi_+1 -> i;
					c;
				enduntil;
				if exacc :XptBoolean separator_ptr then `\n` endif;
				fast_XtFree(text);
			endwhile
		|#) -> string;
		XmStringFreeContext(context);
	else
		false -> string;
	endif;

	text_ptr, context_ptr, separator_ptr,
	repeat 3 times conspair(fixed_ptrs) -> fixed_ptrs endrepeat;
enddefine;


define lconstant Pop_to_motif_string(string) with_props XpmImportString;
	lvars string;
	if string.isstring then
		;;; automatic Xm string creation for normal Pop-11 strings
		XmStringCreateLtoR(string, XmSTRING_DEFAULT_CHARSET) -> string;
	endif;
	string -> XptImportAny("XmString");
enddefine;

Pop_to_motif_string -> updater(XpmCoerceString);
Pop_to_motif_string -> updater(XpmImportString);

;;; XpmCopiedString typespec - this typespec will (1) free the XmString it gets
;;; after coercing it into a pop string and (2) free the XmString it creates
;;; from a pop string when that pop string becomes garbage. The typespec is
;;; used for XmNlabelString resources, since these always copy the xmstring
;;; that the application passes in/gets out.

define XpmCoerceCopiedString(xmstring) -> string;
	lvars xmstring, string;
	if xmstring.isstring then
		xmstring -> string;
	else
		if XpmCoerceString(xmstring) ->> string then
			XmStringFree(xmstring);
		endif;
	endif;
enddefine;
;;;
define updaterof XpmCoerceCopiedString(string) -> xmstring;
	lvars string, xmstring;
	lconstant destroy_action = clearproperty(dup(copy(sys_destroy_action)));
	define lconstant free with_props false with_nargs 2;
		XmStringFree(/*xmstring*/);
		erase(/*string*/);
	enddefine;
	if destroy_action(string) ->> xmstring then
		;;; seen this one before: we can't have more than one destroy
		;;; action, and it saves creating another XmString anyway
		frozval(1, xmstring) -> xmstring;
	else
		->XpmCoerceString(string) -> xmstring;
		;;; when the string becomes garbage, free the XmString
		free(%xmstring%) -> destroy_action(string);
	endif;
enddefine;

endexload_batch;
endsection;


/* --- Revision History ---------------------------------------------------
--- Robert John Duncan, Jun 22 1995
		Fix for updater of XpmCoerceCopiedString for the case where it's
		applied more than once to the same string
--- John Gibson, Apr 14 1993
		Code part of XmConstants to this file, macros and typespecs to
		XmConstants.ph
--- Robert John Duncan, Jul  6 1992
		Changed forward declaration of -XmStringInitContext- to "vars"
		to match the default declaration in "MotifWidgetSet.p"
--- Adrian Howard, Jun 23 1992
		JonM's fixes fixed so they would compile :-)
--- Adrian Howard, Jun 23 1992
		JonM's fixes for XpmCoerceString and XpmCoerceCopiedString installed
--- Jonathan Meyer, Jun  16 1992
	Added XpmCoerceCopiedString. Fixed bug in XpmCoerceString so that it
	frees the C ascii string after use.
--- Adrian Howard, Nov  1 1991 :
		o typespec:XmAnyCallbackStruct renamed typespec:XpmAnyCallbackStruct,
		  fields in typespec renamed as follows:
			reason --> XpmACSReason
			event  --> XpmACSEvent
		o Added padding field to typespec:XpmAnyCallbackStruct
		o XpmAnyCallbackStruct shadowclass renamed XpmAnyCallbackStructPtr
		o Removed XpmACSPad field from XpmAnyCallbackStructPtr
--- Jonathan Meyer, Jul  3 1991
		Moved XpmCoerceString into this file. Added XptLoadClassProc call.
		Defined XmConstants.
--- Jonathan Meyer, May 25 1991
		Added XpmImportString to typespec for XmStrings
--- Robert John Duncan, Mar 21 1991
		Added missing value field to FileSelectionBoxCallbackStruct
--- Jonathan Meyer, Feb 15 1991 Fixed shadowclass structures to follow
		Xpt naming conventions
 */
